"""The Vertica dialect.

https://docs.vertica.com/latest/en/
"""

from sqlfluff.core.dialects import load_raw_dialect
from sqlfluff.core.parser import (
    AnyNumberOf,
    AnySetOf,
    BaseSegment,
    Bracketed,
    BracketedSegment,
    CodeSegment,
    CompositeComparisonOperatorSegment,
    Dedent,
    Delimited,
    IdentifierSegment,
    Indent,
    KeywordSegment,
    LiteralSegment,
    Matchable,
    MultiStringParser,
    OneOf,
    OptionallyBracketed,
    ParseMode,
    Ref,
    RegexLexer,
    RegexParser,
    SegmentGenerator,
    Sequence,
    StringLexer,
    StringParser,
    SymbolSegment,
    TypedParser,
    WordSegment,
)
from sqlfluff.dialects import dialect_ansi as ansi
from sqlfluff.dialects.dialect_vertica_keywords import (
    vertica_reserved_keywords,
    vertica_unreserved_keywords,
)

ansi_dialect = load_raw_dialect("ansi")
vertica_dialect = ansi_dialect.copy_as(
    "vertica",
    formatted_name="Vertica",
    docstring="""The dialect for
`Vertica <https://www.vertica.com/documentation/vertica/all/>`_.""",
)

vertica_dialect.insert_lexer_matchers(
    # Allow ::! operator as in
    # https://docs.vertica.com/latest/en/sql-reference/language-elements/operators/data-type-coercion-operators-cast/cast-failures/
    [
        StringLexer("null_casting_operator", "::!", CodeSegment),
    ],
    before="casting_operator",
)

vertica_dialect.insert_lexer_matchers(
    # Allow <==> operator as in
    # https://docs.vertica.com/latest/en/sql-reference/language-elements/operators/comparison-operators/
    [
        StringLexer("null_equals_operator", "<=>", CodeSegment),
    ],
    before="less_than",
)

vertica_dialect.insert_lexer_matchers(
    # Allow additional math operators as in
    # https://docs.vertica.com/latest/en/sql-reference/language-elements/operators/mathematical-operators/
    # TODO: add other math operators
    [
        StringLexer("integer_division", "//", CodeSegment),
    ],
    before="divide",
)

vertica_dialect.insert_lexer_matchers(
    [
        StringLexer("lambda_arrow", "->", CodeSegment),
        # This is similar to the Unicode regex, the key differences being:
        # - [eE] - must start with e or E
        # - The final quote character must be preceded by:
        # (?<!\\)(?:\\\\)*(?<!')(?:'')     An even/zero number of \ followed by an
        # even/zero number of '
        # OR
        # (?<!\\)(?:\\\\)*\\(?<!')(?:'')*' An odd number of \ followed by an odd number
        # of '
        # There is no UESCAPE block
        RegexLexer(
            "escaped_single_quote",
            r"(?s)[eE](('')+?(?!')|'.*?((?<!\\)(?:\\\\)*"
            r"(?<!')(?:'')*|(?<!\\)(?:\\\\)*\\"
            r"(?<!')(?:'')*')'(?!'))",
            CodeSegment,
        ),
    ],
    before="like_operator",
)

vertica_dialect.patch_lexer_matchers(
    [
        RegexLexer(
            "double_quote",
            r'"([^"]|"")*"',
            CodeSegment,
            segment_kwargs={
                "quoted_value": (r'"((?:[^"]|"")*)"', 1),
                "escape_replacements": [(r'""', '"')],
            },
        ),
        RegexLexer(
            "word",
            r"[\p{L}_][\p{L}\p{N}_$]*",
            WordSegment,
        ),
    ]
)

# Set Keywords
vertica_dialect.update_keywords_set_from_multiline_string(
    "unreserved_keywords", vertica_unreserved_keywords
)

vertica_dialect.update_keywords_set_from_multiline_string(
    "reserved_keywords", vertica_reserved_keywords
)
vertica_dialect.sets("reserved_keywords").difference_update(["ROWS"])

vertica_dialect.sets("bare_functions").clear()
vertica_dialect.sets("bare_functions").update(
    [
        "CURRENT_TIMESTAMP",
        "CURRENT_TIME",
        "CURRENT_DATE",
        "LOCALTIME",
        "LOCALTIMESTAMP",
        "SYSDATE",
    ]
)

# Add all Vertica encoding types
vertica_dialect.sets("encoding_types").clear()
vertica_dialect.sets("encoding_types").update(
    [
        "AUTO",
        "BLOCK_DICT",
        "BLOCKDICT_COMP",
        "BZIP_COMP",
        "COMMONDELTA_COMP",
        "DELTARANGE_COMP",
        "DELTAVAL",
        "GCDDELTA",
        "GZIP_COMP",
        "RLE",
        "ZSTD_COMP",
        "ZSTD_FAST_COMP",
        "ZSTD_HIGH_COMP",
    ],
)

# Add all Vertica compression types (for COPY statement)
vertica_dialect.sets("compression_types").clear()
vertica_dialect.sets("compression_types").update(
    [
        "UNCOMPRESSED",
        "BZIP",
        "GZIP",
        "LZO",
        "ZSTD",
    ],
)

vertica_dialect.sets("date_part_function_name").update(
    ["DATEADD", "DATEDIFF", "EXTRACT", "DATE_PART"]
)

# Add datetime units
# https://docs.vertica.com/latest/en/sql-reference/functions/data-type-specific-functions/datetime-functions/date-part/
vertica_dialect.sets("datetime_units").update(
    [
        "MILLENNIUM",
        "CENTURY",
        "DECADE",
        "EPOCH",
        "YEAR",
        "ISOYEAR",
        "QUARTER",
        "MONTH",
        "WEEK",
        "ISOWEEK",
        "ISODOW",
        "DOW",
        "DOY",
        "DAY",
        "HOUR",
        "MINUTE",
        "SECOND",
        "MILLISECONDS",
        "MICROSECONDS",
        "TIME ZONE",
        "TIMEZONE_HOUR",
        "TIMEZONE_MINUTE",
    ]
)

vertica_dialect.add(
    EncodingType=Sequence(
        MultiStringParser(
            vertica_dialect.sets("encoding_types"),
            KeywordSegment,
            type="encoding_type",
        ),
    ),
    CompressionType=Sequence(
        MultiStringParser(
            vertica_dialect.sets("compression_types"),
            KeywordSegment,
            type="compression_type",
        )
    ),
    IntegerSegment=RegexParser(
        # An unquoted integer that can be passed as an argument to Snowflake functions.
        r"[0-9]+",
        LiteralSegment,
        type="integer_literal",
    ),
    NullCastOperatorSegment=StringParser(
        "::!", SymbolSegment, type="null_casting_operator"
    ),
    NullEqualsOperatorSegment=StringParser(
        "<=>", SymbolSegment, type="null_equals_operator"
    ),
    IntervalUnitsGrammar=OneOf("YEAR", "MONTH", "DAY", "HOUR", "MINUTE", "SECOND"),
    InterpolateGrammar=Sequence("INTERPOLATE", OneOf("PREVIOUS", "NEXT"), "VALUE"),
    IntervalLiteralGrammar=Sequence(
        Ref("IntervalUnitsGrammar"),
        Sequence(
            "TO",
            Sequence(
                Ref("IntervalUnitsGrammar"),
                Ref("BracketedArguments", optional=True),
            ),
            optional=True,
        ),
    ),
    IntegerDivideSegment=StringParser("//", SymbolSegment, type="binary_operator"),
    LambdaArrowSegment=StringParser("->", SymbolSegment, type="lambda_arrow"),
)

vertica_dialect.replace(
    FunctionContentsExpressionGrammar=OneOf(
        Ref("LambdaExpressionSegment"),
        Ref("ExpressionSegment"),
    ),
    FunctionContentsGrammar=AnyNumberOf(
        Ref("ExpressionSegment"),
        OptionallyBracketed(Ref("SetExpressionSegment")),
        # A Cast-like function
        Sequence(
            Ref("ExpressionSegment"),
            "AS",
            OneOf(Ref("DatatypeSegment"), Ref("DateTimeLiteralGrammar")),
        ),
        # Trim function
        Sequence(
            Ref("TrimParametersGrammar"),
            Ref("ExpressionSegment", optional=True, exclude=Ref.keyword("FROM")),
            "FROM",
            Ref("ExpressionSegment"),
        ),
        # An extract-like or substring-like function
        # https://www.postgresql.org/docs/current/functions-string.html
        Sequence(
            OneOf(Ref("DatetimeUnitSegment"), Ref("ExpressionSegment")),
            AnySetOf(
                Sequence("FROM", Ref("ExpressionSegment")),
                Sequence("FOR", Ref("ExpressionSegment")),
                optional=True,
            ),
        ),
        Sequence(
            # Allow an optional distinct keyword here.
            Ref.keyword("DISTINCT", optional=True),
            OneOf(
                # Most functions will be using the delimited route
                # but for COUNT(*) or similar we allow the star segment
                # here.
                Ref("StarSegment"),
                Delimited(Ref("FunctionContentsExpressionGrammar")),
            ),
        ),
        Ref(
            "AggregateOrderByClause"
        ),  # used by string_agg (postgres), group_concat (exasol),listagg (snowflake)..
        Sequence(Ref.keyword("SEPARATOR"), Ref("LiteralGrammar")),
        # like a function call: POSITION ( 'QL' IN 'SQL')
        Sequence(
            Ref("DatatypeSegment", optional=True),
            OneOf(
                Ref("QuotedLiteralSegment"),
                Ref("SingleIdentifierGrammar"),
                Ref("ColumnReferenceSegment"),
            ),
            "IN",
            Ref("DatatypeSegment", optional=True),
            OneOf(
                Ref("QuotedLiteralSegment"),
                Ref("SingleIdentifierGrammar"),
                Ref("ColumnReferenceSegment"),
            ),
        ),
        # used by listagg, explode, unnest
        Sequence(
            Ref.keyword("DISTINCT", optional=True),
            OneOf(
                Ref("QuotedLiteralSegment"),
                Ref("ColumnReferenceSegment"),
                Ref("ExpressionSegment"),
            ),
            Sequence(
                "USING",
                "PARAMETERS",
                Delimited(
                    Sequence(
                        Ref("ParameterNameSegment"),
                        Ref("EqualsSegment"),
                        OneOf(
                            Ref("QuotedLiteralSegment"),
                            Ref("BooleanLiteralGrammar"),
                            Ref("NumericLiteralSegment"),
                        ),
                    ),
                ),
            ),
            Ref("OverClauseSegment", optional=True),
        ),
        Ref("IgnoreRespectNullsGrammar"),
        Ref("EmptyStructLiteralSegment"),
    ),
    ObjectReferenceTerminatorGrammar=OneOf(
        "ON",
        "AS",
        "USING",
        Ref("CommaSegment"),
        Ref("CastOperatorSegment"),
        Ref("NullCastOperatorSegment"),
        Ref("StartSquareBracketSegment"),
        Ref("StartBracketSegment"),
        Ref("BinaryOperatorGrammar"),
        Ref("ColonSegment"),
        Ref("DelimiterGrammar"),
        Ref("JoinLikeClauseGrammar"),
        BracketedSegment,
    ),
    PostFunctionGrammar=AnySetOf(
        # Optional OVER suffix for window functions.
        # This is supported in bigquery & postgres (and its derivatives)
        # and so is included here for now.
        Ref("OverClauseSegment"),
        # Filter clause supported by both Postgres and SQLite
        Ref("FilterClauseGrammar"),
        # Within group clause supported by some analytic functions in Vertica
        Ref("WithinGroupClauseSegment"),
    ),
    DateTimeLiteralGrammar=Sequence(
        # analog of postgres dialect but with treating expressions like
        # as interval hour TO SECOND(6)
        OneOf("DATE", "TIME", "TIMESTAMP", "INTERVAL"),
        TypedParser(
            "single_quote",
            LiteralSegment,
            type="date_constructor_literal",
            optional=True,
        ),
        Ref("IntervalLiteralGrammar", optional=True),
    ),
    Expression_A_Grammar=Sequence(
        # It's a copy of ansi Expression_A_Grammar
        Ref("Tail_Recurse_Expression_A_Grammar"),
        AnyNumberOf(
            OneOf(
                Ref("LikeExpressionGrammar"),
                Sequence(
                    Ref("BinaryOperatorGrammar"),
                    Ref("Tail_Recurse_Expression_A_Grammar"),
                ),
                Ref("InOperatorGrammar"),
                Sequence(
                    "IS",
                    Ref.keyword("NOT", optional=True),
                    Ref("IsClauseGrammar"),
                ),
                Ref("IsNullGrammar"),
                Sequence(
                    Sequence(
                        Ref("InterpolateGrammar"),
                    ),
                    Ref("Expression_A_Grammar"),
                ),
                Ref("NotNullGrammar"),
                Ref("CollateGrammar"),
                Sequence(
                    Ref.keyword("NOT", optional=True),
                    "BETWEEN",
                    Ref("Expression_B_Grammar"),
                    "AND",
                    Ref("Tail_Recurse_Expression_A_Grammar"),
                ),
            )
        ),
    ),
    ComparisonOperatorGrammar=OneOf(
        Ref("EqualsSegment"),
        Ref("NullEqualsSegment"),
        Ref("GreaterThanSegment"),
        Ref("LessThanSegment"),
        Ref("GreaterThanOrEqualToSegment"),
        Ref("LessThanOrEqualToSegment"),
        Ref("NotEqualToSegment"),
        Ref("LikeOperatorSegment"),
        Ref("IsDistinctFromGrammar"),
    ),
    JoinTypeKeywordsGrammar=OneOf(
        "ANTI",
        "SEMIALL",
        "SEMI",
        Sequence("NULLAWARE", "ANTI"),
        "INNER",
        Sequence(
            OneOf(
                "FULL",
                "LEFT",
                "RIGHT",
            ),
            Ref.keyword("OUTER", optional=True),
        ),
        Sequence(
            "NATURAL",
            OneOf(
                "INNER",
                Sequence(OneOf("RIGHT", "LEFT", "FULL"), "OUTER"),
            ),
        ),
    ),
    ArithmeticBinaryOperatorGrammar=OneOf(
        Ref("PlusSegment"),
        Ref("MinusSegment"),
        Ref("DivideSegment"),
        Ref("IntegerDivideSegment"),
        Ref("MultiplySegment"),
        Ref("ModuloSegment"),
        Ref("BitwiseAndSegment"),
        Ref("BitwiseOrSegment"),
        Ref("BitwiseXorSegment"),
        Ref("BitwiseLShiftSegment"),
        Ref("BitwiseRShiftSegment"),
    ),
    # Vertica supports the non-standard ISNULL and NONNULL comparison operators. See
    # https://docs.vertica.com/latest/en/sql-reference/language-elements/operators/null-operators/
    IsNullGrammar=Ref.keyword("ISNULL"),
    NotNullGrammar=Ref.keyword("NOTNULL"),
    QuotedLiteralSegment=OneOf(
        Sequence(
            TypedParser(
                "single_quote",
                LiteralSegment,
                type="quoted_literal",
            ),
        ),
        # Support Extended string literals
        Sequence(
            TypedParser(
                "escaped_single_quote",
                LiteralSegment,
                type="quoted_literal",
            ),
        ),
    ),
    QuotedIdentifierSegment=TypedParser(
        "double_quote", IdentifierSegment, type="quoted_identifier", casefold=str.upper
    ),
    NakedIdentifierSegment=SegmentGenerator(
        # Generate the anti template from the set of reserved keywords
        lambda dialect: RegexParser(
            # https://docs.vertica.com/24.3.x/en/sql-reference/language-elements/identifiers/
            # Unquoted SQL identifiers must begin with one of the following:
            # * Non-Unicode letters: A–Z or a-z
            # -- /actually Vertica accepts also non-ASCII UTF-8 Unicode
            # characters here, which is not well documented/
            # * Underscore (_)
            # Subsequent characters in an identifier can be any combination of
            # the following:
            # * Non-Unicode letters: A–Z or a-z
            # * Underscore (_)
            # * Digits(0–9)
            # * Unicode letters (letters with diacriticals or not in the Latin
            # alphabet), unsupported for model names
            # * Dollar sign ($), unsupported for model names
            #
            # Vertica accepts **non-ASCII UTF-8 Unicode characters** for table
            # names, column names, and other identifiers,
            # extending the cases where upper/lower case distinctions are
            # ignored (case-folded) to all alphabets,
            # including Latin, Cyrillic, and Greek.
            # \p{L} matches any kind of letter from any language;
            # \p{N} matches any kind of numeric character in any script
            r"[\p{L}_][\p{L}\p{N}$_]*",
            IdentifierSegment,
            type="naked_identifier",
            anti_template=r"^("
            + r"|".join(sorted(dialect.sets("reserved_keywords")))
            + r")$",
            casefold=str.upper,
        )
    ),
    ParameterNameSegment=RegexParser(
        # need to cover cases where non-ascii word is parameter
        # like ```ALTER TABLE some_table TO utf8_identifier_eg_Verkäufer;```
        r"[\p{L}_][\p{L}\p{N}$_]*",
        CodeSegment,
        type="parameter",
    ),
)


class ShorthandCastSegment(ansi.ShorthandCastSegment):
    """A casting operation using '::' or '::!'."""

    match_grammar: Matchable = Sequence(
        OneOf(
            Ref("Expression_D_Grammar"),
            Ref("CaseExpressionSegment"),
        ),
        AnyNumberOf(
            Sequence(
                OneOf(Ref("CastOperatorSegment"), Ref("NullCastOperatorSegment")),
                Ref("DatatypeSegment"),
                Ref("TimeZoneGrammar", optional=True),
                allow_gaps=True,
            ),
            min_times=1,
        ),
    )


class StatementSegment(ansi.StatementSegment):
    """A generic segment, to any of its child subsegments."""

    match_grammar = ansi.StatementSegment.match_grammar.copy(
        insert=[
            Ref("CreateExternalTableSegment"),
            Ref("CreateTableLikeStatementSegment"),
            Ref("CreateTableAsStatementSegment"),
            Ref("CreateProjectionStatementSegment"),
            Ref("AlterDefaultPrivilegesGrantSegment"),
            Ref("DropProjectionStatementSegment"),
            Ref("AlterViewStatementSegment"),
            Ref("SetStatementSegment"),
            Ref("CommentOnStatementSegment"),
            Ref("TransactionalStatements"),
            Ref("AlterSessionStatements"),
            Ref("CopyStatementSegment"),
            Ref("AlterSchemaStatementSegment"),
        ],
    )


class ArrayTypeSegment(ansi.ArrayTypeSegment):
    """Prefix for array literals specifying the type."""

    type = "array_type"
    match_grammar = Sequence(
        Ref.keyword("ARRAY"),
        Bracketed(
            Ref("DatatypeSegment"),
            bracket_type="square",
            bracket_pairs_set="bracket_pairs",
            optional=True,
        ),
    )


class LimitClauseSegment(ansi.LimitClauseSegment):
    """A vertica `LIMIT` clause.

    https://docs.vertica.com/latest/en/sql-reference/statements/select/limit-clause/
    """

    match_grammar: Matchable = Sequence(
        "LIMIT",
        Indent,
        OptionallyBracketed(
            OneOf(
                # Allow a number by itself OR
                Ref("NumericLiteralSegment"),
                # An arbitrary expression
                Ref("ExpressionSegment"),
                "ALL",
            )
        ),
        OneOf(
            Sequence(
                "OFFSET",
                OneOf(
                    # Allow a number by itself OR
                    Ref("NumericLiteralSegment"),
                    # An arbitrary expression
                    Ref("ExpressionSegment"),
                ),
            ),
            Sequence(
                Ref("CommaSegment"),
                Ref("NumericLiteralSegment"),
            ),
            Ref("OverClauseSegment"),
            optional=True,
        ),
        Dedent,
    )


class ColumnEncodingSegment(BaseSegment):
    """The `ENCODING` clause within a `CREATE TABLE` statement for a column."""

    type = "column_encoding"
    match_grammar: Matchable = Sequence(
        "ENCODING",
        Ref("EncodingType"),
    )


class TableConstraintSegment(ansi.TableConstraintSegment):
    """A table constraint, e.g. for CREATE TABLE.

    As specified in
    https://docs.vertica.com/latest/en/sql-reference/statements/create-statements/create-table/table-constraint/
    """

    match_grammar = Sequence(
        Sequence(  # [ CONSTRAINT <Constraint name> ]
            "CONSTRAINT", Ref("ObjectReferenceSegment"), optional=True
        ),
        OneOf(
            Sequence(  # PRIMARY KEY (column[,...]) [ ENABLED | DISABLED]
                Ref("PrimaryKeyGrammar"),
                # Columns making up PRIMARY KEY constraint
                Ref("BracketedColumnReferenceListGrammar"),
                OneOf("ENABLED", "DISABLED", optional=True),
            ),
            Sequence(
                "CHECK",
                Bracketed(Ref("ExpressionSegment")),
                OneOf("ENABLED", "DISABLED", optional=True),
            ),
            Sequence(  # UNIQUE (column[,...]) [ENABLED | DISABLED]
                "UNIQUE",
                Ref("BracketedColumnReferenceListGrammar"),
                OneOf("ENABLED", "DISABLED", optional=True),
            ),
            Sequence(  # FOREIGN KEY ( column_name [, ... ] )
                # REFERENCES reftable [ ( refcolumn [, ... ] ) ]
                "FOREIGN",
                "KEY",
                # Local columns making up FOREIGN KEY constraint
                Ref("BracketedColumnReferenceListGrammar"),
                Ref(
                    "ReferenceDefinitionGrammar"
                ),  # REFERENCES reftable [ ( refcolumn) ]
            ),
        ),
    )


class LikeOptionSegment(BaseSegment):
    """Like Option Segment.

    As specified in
    https://docs.vertica.com/latest/en/admin/working-with-native-tables/creating-table-from-other-tables/replicating-table/
    """

    type = "like_option_segment"

    match_grammar = Sequence(
        OneOf(
            Sequence(OneOf("INCLUDING", "EXCLUDING"), "PROJECTIONS"),
            Ref("SchemaPrivilegesSegment"),
        ),
    )


class DiskQuotaSegment(BaseSegment):
    """Disk Quota Segment.

    https://docs.vertica.com/latest/en/admin/working-with-native-tables/disk-quotas/
    Available from Vertica 12.x
    """

    type = "disk_quota_segment"

    match_grammar = Sequence("DISK_QUOTA", Ref("QuotedLiteralSegment"))


class KsafeSegment(BaseSegment):
    """Ksafe Segment.

    https://docs.vertica.com/latest/en/sql-reference/statements/create-statements/create-table/
    https://docs.vertica.com/latest/en/architecture/enterprise-concepts/k-safety-an-enterprise-db/
    """

    type = "ksafe_segment"

    match_grammar = Sequence(
        "KSAFE",
        Ref("NumericLiteralSegment", optional=True),
    )


class SchemaPrivilegesSegment(BaseSegment):
    """Schema Privileges Segment.

    https://docs.vertica.com/latest/en/sql-reference/statements/create-statements/create-table/
    """

    type = "schema_privileges_segment"
    match_grammar: Matchable = Sequence(
        # MATERIALIZE available only in ALTER TABLE statement,
        # but we keep it here to not duplicate the code
        OneOf("INCLUDE", "EXCLUDE", "MATERIALIZE"),
        Ref.keyword("SCHEMA", optional=True),
        "PRIVILEGES",
    )


class SegmentedByClauseSegment(BaseSegment):
    """A `SEGMENTED BY` or `UNSEGMENTED` clause.

    As specified in
    https://docs.vertica.com/latest/en/sql-reference/statements/
    create-statements/create-projection/hash-segmentation-clause/
    Vertica allows different expressions in segmented by clause,
    but using hash function is recommended one
    As specified in
    https://docs.vertica.com/latest/en/sql-reference/statements/
    create-statements/create-projection/unsegmented-clause/
    """

    type = "segmentedby_clause"
    match_grammar: Matchable = Sequence(
        OneOf(
            Sequence("UNSEGMENTED", "ALL", "NODES"),
            Sequence(
                "SEGMENTED",
                "BY",
                OneOf(
                    Ref("FunctionSegment"),
                    Bracketed(
                        Delimited(
                            Sequence(
                                OneOf(
                                    Ref("ColumnReferenceSegment"),
                                    Ref("NumericLiteralSegment"),
                                    Ref("ExpressionSegment"),
                                    Ref("ShorthandCastSegment"),
                                ),
                            ),
                        ),
                    ),
                ),
                "ALL",
                "NODES",
            ),
        ),
    )


class PartitionByClauseSegment(BaseSegment):
    """A `PARTITION BY` clause.

    As specified in
    https://docs.vertica.com/latest/en/sql-reference/statements/create-statements/create-table/partition-clause/
    """

    type = "partitionby_clause"
    match_grammar: Matchable = Sequence(
        "PARTITION",
        "BY",
        AnyNumberOf(
            Delimited(
                Sequence(
                    AnyNumberOf(
                        Ref("ColumnReferenceSegment"),
                        Ref("ExpressionSegment"),
                        Ref("FunctionSegment"),
                        Ref("ShorthandCastSegment"),
                    ),
                ),
            ),
            Bracketed(
                Delimited(
                    Sequence(
                        AnyNumberOf(
                            Ref("ColumnReferenceSegment"),
                            Ref("FunctionSegment"),
                            Ref("ShorthandCastSegment"),
                        ),
                    ),
                ),
            ),
            terminators=[Sequence("GROUP", "BY"), "REORGANIZE"],
        ),
        Ref("GroupByClauseSegment", optional=True),
        Ref.keyword("REORGANIZE", optional=True),
    )


class CreateTableStatementSegment(ansi.CreateTableStatementSegment):
    """A `CREATE TABLE` statement.

    https://docs.vertica.com/latest/en/sql-reference/statements/create-statements/create-table/
    """

    match_grammar = Sequence(
        "CREATE",
        OneOf(
            Sequence(
                OneOf("GLOBAL", "LOCAL", optional=True),
                Ref("TemporaryGrammar", optional=True),
            ),
            optional=True,
        ),
        "TABLE",
        Ref("IfNotExistsGrammar", optional=True),
        Ref("TableReferenceSegment"),
        Sequence(
            Bracketed(
                Delimited(
                    Sequence(
                        Ref("ColumnReferenceSegment"),
                        Ref("DatatypeSegment"),
                        AnyNumberOf(
                            Ref("ColumnConstraintSegment"),
                            Ref("ColumnEncodingSegment"),
                            Sequence(
                                "ACCESSRANK", Ref("IntegerSegment"), optional=True
                            ),
                        ),
                    ),
                    Ref("TableConstraintSegment"),
                ),
            ),
        ),
        AnySetOf(
            Ref("OrderByClauseSegment"),
            Ref("SegmentedByClauseSegment"),
            Ref("KsafeSegment"),
            Ref("SchemaPrivilegesSegment"),
            Ref("DiskQuotaSegment"),
            Ref("PartitionByClauseSegment"),
        ),
        AnySetOf(
            # these options are available only for temp table, so it's kind of a hack
            Sequence("ON", "COMMIT", OneOf("DELETE", "PRESERVE"), "ROWS"),
            Sequence("NO", "PROJECTION"),
        ),
    )


class CreateTableAsStatementSegment(BaseSegment):
    """A `CREATE TABLE AS` statement.

    As specified in
    https://docs.vertica.com/latest/en/admin/working-with-native-tables/creating-table-from-other-tables/creating-table-from-query/
    """

    type = "create_table_as_statement"

    match_grammar = Sequence(
        "CREATE",
        OneOf(
            Sequence(
                OneOf("GLOBAL", "LOCAL", optional=True),
                Ref("TemporaryGrammar", optional=True),
            ),
            optional=True,
        ),
        "TABLE",
        Ref("IfNotExistsGrammar", optional=True),
        Ref("TableReferenceSegment"),
        AnySetOf(
            # these options are available only for temp table, so it's kind of a hack
            Sequence("ON", "COMMIT", OneOf("DELETE", "PRESERVE"), "ROWS"),
            Sequence("NO", "PROJECTION"),
        ),
        AnyNumberOf(
            Bracketed(
                Delimited(
                    Sequence(
                        Ref("ColumnReferenceSegment"),
                        Ref("ColumnEncodingSegment", optional=True),
                        Sequence("ACCESSRANK", Ref("IntegerSegment"), optional=True),
                        # TODO: need to add GROUPED clause
                        # https://docs.vertica.com/latest/en/sql-reference/statements/
                        # create-statements/create-projection/grouped-clause/
                    ),
                ),
                optional=True,
            ),
            Ref("SchemaPrivilegesSegment", optional=True),
        ),
        "AS",
        # TODO: need to add LABEL clause
        # https://docs.vertica.com/latest/en/admin/
        # working-with-native-tables/creating-table-from-other-tables/creating-table-from-query/
        Sequence(
            "AT",
            OneOf("LATEST", Ref("NumericLiteralSegment"), Ref("DatetimeUnitSegment")),
            optional=True,
        ),
        Ref(
            "SelectableGrammar",
            terminators=[Ref("SegmentedByClauseSegment"), Ref("OrderByClauseSegment")],
        ),
        Ref("OrderByClauseSegment", optional=True),
        Ref("SegmentedByClauseSegment", optional=True),
    )


class CreateTableLikeStatementSegment(BaseSegment):
    """A `CREATE TABLE LIKE` statement.

    As specified in
    https://docs.vertica.com/latest/en/admin/working-with-native-tables/creating-table-from-other-tables/replicating-table/
    """

    type = "create_table_like_statement"

    match_grammar = Sequence(
        "CREATE",
        "TABLE",
        Ref("IfNotExistsGrammar", optional=True),
        Ref("TableReferenceSegment"),
        Sequence(
            "LIKE",
            Ref("TableReferenceSegment"),
            AnyNumberOf(Ref("LikeOptionSegment"), optional=True),
        ),
        Ref("DiskQuotaSegment", optional=True),
    )


class CopyOptionsForColumnsSegment(BaseSegment):
    """A vertica options for columns in COPY.

    https://docs.vertica.com/latest/en/sql-reference/statements/copy/
    """

    type = "copy_options_for_columns"

    match_grammar = Sequence(
        AnySetOf(
            Sequence(
                "DELIMITER", Sequence("AS", optional=True), Ref("QuotedLiteralSegment")
            ),
            Sequence(
                "ENCLOSED", Sequence("BY", optional=True), Ref("QuotedLiteralSegment")
            ),
            "ENFORCELENGTH",
            OneOf(
                Sequence(
                    "ESCAPE", Sequence("AS", optional=True), Ref("QuotedLiteralSegment")
                ),
                Sequence("NO", "ESCAPE"),
            ),
            Sequence("FILLER", Ref("DatatypeSegment")),
            Sequence("FORMAT", Ref("QuotedLiteralSegment")),
            Sequence(
                "NULL", Sequence("AS", optional=True), Ref("QuotedLiteralSegment")
            ),
            Sequence("TRIM", Ref("QuotedLiteralSegment")),
        ),
    )


class CopyColumnOptionsSegment(BaseSegment):
    """A vertica column description in COPY.

    https://docs.vertica.com/latest/en/sql-reference/statements/copy/
    """

    type = "copy_column_options"

    match_grammar = Sequence(
        Ref("ColumnReferenceSegment"),
        Ref("CopyOptionsForColumnsSegment", optional=True),
    )


class CopyOptionsSegment(BaseSegment):
    """A vertica options for COPY.

    https://docs.vertica.com/latest/en/sql-reference/statements/copy/
    """

    type = "copy_options"

    match_grammar = Sequence(
        AnyNumberOf(
            # TODO: add WITH FILTER, WITH PARSER, and on nodename support
            Sequence("ABORT", "ON", "ERROR"),
            Sequence("ERROR", "TOLERANCE"),
            Sequence("EXCEPTION", Ref("QuotedLiteralSegment")),
            Sequence("RECORD", "TERMINATOR", Ref("QuotedLiteralSegment")),
            Sequence("REJECTED", "DATA", Ref("QuotedLiteralSegment")),
            Sequence("REJECTMAX", Ref("IntegerSegment")),
            Sequence("SKIP", Ref("IntegerSegment")),
            Sequence("SKIP", "BYTES", Ref("IntegerSegment")),
            Sequence("TRAILING", "NULLCOLS"),
            Ref("CopyOptionsForColumnsSegment", optional=True),
        ),
    )


class CreateExternalTableSegment(BaseSegment):
    """A vertica `CREATE EXTERNAL TABLE` statement.

    https://docs.vertica.com/latest/en/sql-reference/statements/create-statements/create-external-table-as-copy/
    """

    type = "create_external_table_statement"

    match_grammar = Sequence(
        "CREATE",
        "EXTERNAL",
        "TABLE",
        Ref("IfNotExistsGrammar", optional=True),
        Ref("TableReferenceSegment"),
        # Columns:
        Sequence(
            Bracketed(
                Delimited(
                    Sequence(
                        Ref("ColumnReferenceSegment"),
                        Ref("DatatypeSegment"),
                        AnyNumberOf(
                            Ref("ColumnConstraintSegment"),
                            Ref("ColumnEncodingSegment"),
                            Sequence(
                                "ACCESSRANK", Ref("IntegerSegment"), optional=True
                            ),
                        ),
                    ),
                    Ref("TableConstraintSegment"),
                ),
            ),
        ),
        Ref("SchemaPrivilegesSegment", optional=True),
        "AS",
        "COPY",
        OneOf(
            Bracketed(
                Delimited(
                    Sequence(
                        Ref("ColumnReferenceSegment"),
                        Ref("CopyColumnOptionsSegment", optional=True),
                    ),
                ),
            ),
            Sequence(
                "COLUMN",
                "OPTION",
                Bracketed(
                    Delimited(
                        Sequence(
                            Ref("ColumnReferenceSegment"),
                            Ref("CopyColumnOptionsSegment", optional=True),
                        ),
                    ),
                ),
            ),
            optional=True,
        ),
        "FROM",
        Ref("QuotedLiteralSegment"),
        OneOf("NATIVE", Sequence("NATIVE", "VARCHAR"), "ORC", "PARQUET", optional=True),
        Ref("CopyOptionsSegment", optional=True),
    )


class GroupByClauseSegment(BaseSegment):
    """A `GROUP BY` clause like in `SELECT`."""

    type = "groupby_clause"

    match_grammar: Matchable = Sequence(
        "GROUP",
        "BY",
        Indent,
        Sequence(
            Delimited(
                OneOf(
                    Ref("ColumnReferenceSegment"),
                    # Can `GROUP BY 1`
                    Ref("NumericLiteralSegment"),
                    # Can `GROUP BY coalesce(col, 1)`
                    Ref("ExpressionSegment"),
                ),
                terminators=[Ref("GroupByClauseTerminatorGrammar")],
            ),
        ),
        Dedent,
    )


class CreateProjectionStatementSegment(BaseSegment):
    """A `CREATE PROJECTION` statement.

    As specified in
    https://docs.vertica.com/latest/en/sql-reference/statements/
    create-statements/create-projection/standard-projection/
    """

    type = "create_projection_statement"

    match_grammar = Sequence(
        "CREATE",
        "PROJECTION",
        Ref("IfNotExistsGrammar", optional=True),
        Ref("TableReferenceSegment"),
        Bracketed(
            Delimited(
                Sequence(
                    Ref("ColumnReferenceSegment"),
                    Ref("ColumnEncodingSegment", optional=True),
                    Sequence("ACCESSRANK", Ref("IntegerSegment"), optional=True),
                    # TODO: need to add GROUPED clause
                    # https://docs.vertica.com/latest/en/sql-reference/statements/
                    # create-statements/create-projection/grouped-clause/
                ),
            ),
            optional=True,
        ),
        "AS",
        Ref(
            "SelectableGrammar",
            terminators=[
                Ref("SegmentedByClauseSegment"),
                Ref("OrderByClauseSegment"),
                Ref("LimitClauseSegment"),
                Ref("GroupByClauseSegment"),
                "ON",
            ],
        ),
        OneOf(
            # TODO: add udtf projection type
            AnyNumberOf(
                Ref("OrderByClauseSegment"),
                Ref("SegmentedByClauseSegment"),
                Sequence(
                    "ON",
                    "PARTITION",
                    "RANGE",
                    "BETWEEN",
                    Ref("QuotedLiteralSegment"),
                    "AND",
                    Ref("QuotedLiteralSegment"),
                ),
            ),
            Ref("GroupByClauseSegment"),
            Ref("LimitClauseSegment"),
        ),
        Ref("KsafeSegment", optional=True),
    )


class AlterTableStatementSegment(ansi.AlterTableStatementSegment):
    """An `ALTER TABLE` statement.

    https://docs.vertica.com/latest/en/sql-reference/statements/alter-statements/alter-table/
    """

    match_grammar = Sequence(
        "ALTER",
        "TABLE",
        Delimited(Ref("TableReferenceSegment")),
        OneOf(
            Sequence(
                Delimited(Ref("AlterTableActionSegment")),
            ),
            Sequence(
                "ADD",
                Ref.keyword("COLUMN"),
                Ref("IfNotExistsGrammar", optional=True),
                Ref("ColumnReferenceSegment"),
                Ref("DatatypeSegment"),
                AnyNumberOf(Ref("ColumnConstraintSegment")),
                Ref("ColumnEncodingSegment", optional=True),
                OneOf(
                    Sequence(
                        "PROJECTIONS",
                        Bracketed(Delimited(Ref("TableReferenceSegment"))),
                    ),
                    Sequence("ALL", "PROJECTIONS"),
                    optional=True,
                ),
                Ref("ColumnConstraintSegment", optional=True),
            ),
            Sequence(
                "ALTER",
                Ref.keyword("COLUMN"),
                Ref("ColumnReferenceSegment"),
                OneOf(
                    Sequence(
                        Ref("ColumnEncodingSegment"),
                        "PROJECTIONS",
                        Bracketed(Delimited(Ref("TableReferenceSegment"))),
                    ),
                    Sequence("SET", Ref("ColumnSetSegment")),
                    Sequence("SET", "NOT", "NULL"),
                    Sequence("SET", "DATA", "TYPE", Ref("DatatypeSegment")),
                    Sequence(
                        "DROP",
                        OneOf(
                            "DEFAULT",
                            Sequence("SET", "USING"),
                            Sequence("DEFAULT", "USING"),
                            Sequence("NOT", "NULL"),
                        ),
                    ),
                ),
            ),
            Sequence(
                "DROP",
                "CONSTRAINT",
                Ref("ParameterNameSegment"),
                OneOf("CASCADE", "RESTRICT", optional=True),
            ),
            Sequence(
                "DROP",
                Ref.keyword("COLUMN", optional=True),
                Ref("IfExistsGrammar", optional=True),
                Ref("ColumnReferenceSegment"),
                Ref("DropBehaviorGrammar", optional=True),
            ),
            Ref("PartitionByClauseSegment"),
            Sequence("REMOVE", "PARTITIONING"),
            Sequence(
                "RENAME",
                Ref.keyword("COLUMN", optional=True),
                Ref("ColumnReferenceSegment"),
                "TO",
                Ref("ColumnReferenceSegment"),
            ),
            Sequence(
                "RENAME",
                "TO",
                Delimited(Ref("TableReferenceSegment")),
            ),
            "REORGANIZE",
            Sequence("SET", "SCHEMA", Ref("SchemaReferenceSegment")),
        ),
    )


class AlterTableActionSegment(BaseSegment):
    """Alter Table Action Segment.

    https://docs.vertica.com/latest/en/sql-reference/statements/alter-statements/alter-table/
    """

    type = "alter_table_action_segment"

    match_grammar = OneOf(
        Sequence("ADD", Ref("TableConstraintSegment")),
        Sequence(
            "ALTER",
            "CONSTRAINT",
            Ref("ParameterNameSegment"),
            OneOf("ENABLED", "DISABLED"),
        ),
        Ref("DiskQuotaSegment"),
        Sequence("FORCE", "OUTER", Ref("IntegerSegment")),
        Ref("SchemaPrivilegesSegment"),
        Sequence(
            "OWNER",
            "TO",
            Ref("ParameterNameSegment"),
        ),
        Sequence(
            "SET",
            OneOf(
                Sequence(
                    "ActivePartitionCount", OneOf(Ref("IntegerSegment"), "DEFAULT")
                ),
                Sequence("IMMUTABLE", "ROWS"),
                Sequence("MERGEOUT", OneOf("1", "2")),
            ),
        ),
    )


class AlterDefaultPrivilegesObjectPrivilegesSegment(BaseSegment):
    """`ALTER DEFAULT PRIVILEGES` object privileges.

    https://docs.vertica.com/latest/en/sql-reference/statements/grant-statements/grant-table/
    """

    type = "alter_default_privileges_object_privilege"
    match_grammar = OneOf(
        Sequence(
            "ALL",
            Ref.keyword("PRIVILEGES", optional=True),
            Ref.keyword("EXTEND", optional=True),
        ),
        Delimited(
            "SELECT",
            "INSERT",
            "UPDATE",
            "DELETE",
            "REFERENCES",
            "TRUNCATE",
            "ALTER",
            "DROP",
            terminators=["ON"],
        ),
    )


class AlterDefaultPrivilegesGrantSegment(BaseSegment):
    """`GRANT` for `ALTER DEFAULT PRIVILEGES`.

    https://docs.vertica.com/latest/en/sql-reference/statements/grant-statements/grant-table/
    """

    type = "alter_default_privileges_grant"
    match_grammar = Sequence(
        "GRANT",
        Ref("AlterDefaultPrivilegesObjectPrivilegesSegment"),
        "ON",
        OneOf(
            Delimited(
                Sequence(
                    Ref.keyword("TABLE", optional=True), Ref("TableReferenceSegment")
                )
            ),
            Delimited(
                Sequence(
                    "ALL", "TABLES", "IN", "SCHEMA", Ref("SchemaReferenceSegment")
                ),
            ),
            terminators=["WITH"],
        ),
        "TO",
        Delimited(
            Ref("RoleReferenceSegment"),
            terminators=["WITH"],
        ),
        Sequence("WITH", "GRANT", "OPTION", optional=True),
    )


class ColumnConstraintSegment(ansi.ColumnConstraintSegment):
    """A column option; each CREATE TABLE column can have 0 or more.

    https://docs.vertica.com/latest/en/sql-reference/statements/create-statements/create-table/column-constraint/
    """

    match_grammar = Sequence(
        # TODO: add auto increment
        OneOf(
            Sequence(
                Sequence(
                    "CONSTRAINT",
                    Ref("ObjectReferenceSegment"),  # Constraint name
                    optional=True,
                ),
                OneOf(
                    Sequence(
                        Ref.keyword("NOT", optional=True), "NULL"
                    ),  # NOT NULL or NULL
                    Sequence(
                        "CHECK",
                        Bracketed(Ref("ExpressionSegment")),
                        OneOf("ENABLED", "DISABLED", optional=True),
                    ),
                    Sequence(
                        "UNIQUE",
                        OneOf("ENABLED", "DISABLED", optional=True),
                    ),
                    Sequence(
                        "PRIMARY",
                        "KEY",
                        OneOf("ENABLED", "DISABLED", optional=True),
                    ),
                    # REFERENCES reftable [ ( refcolumn) ]
                    Ref("ReferenceDefinitionGrammar"),
                ),
            ),
            Ref("ColumnSetSegment"),
        ),
    )


class ColumnSetSegment(BaseSegment):
    """A SET DEFAULT | USING | DEFAULT USING.

    https://docs.vertica.com/latest/en/sql-reference/statements/alter-statements/alter-table/
    """

    type = "column_set_segment"

    match_grammar = Sequence(  # DEFAULT <value>
        OneOf(
            "DEFAULT",
            # Depends on the place where we use it (in alter table or create table)
            # we need or don't need set keyword
            Sequence(Ref.keyword("SET", optional=True), "USING"),
            Sequence("DEFAULT", "USING"),
        ),
        OneOf(
            Ref("ShorthandCastSegment"),
            Ref("LiteralGrammar"),
            Ref("FunctionSegment"),
            Ref("BareFunctionSegment"),
            Ref("ExpressionSegment"),
            Bracketed(Ref("SelectableGrammar")),
        ),
    )


class DropProjectionStatementSegment(BaseSegment):
    """A `DROP PROJECTION` statement.

    https://docs.vertica.com/latest/en/sql-reference/statements/drop-statements/drop-projection/
    """

    type = "drop_projection_statement"

    match_grammar: Matchable = Sequence(
        "DROP",
        "PROJECTION",
        Ref("IfExistsGrammar", optional=True),
        Delimited(Ref("TableReferenceSegment")),
        Ref("DropBehaviorGrammar", optional=True),
    )


class AlterViewStatementSegment(BaseSegment):
    """A `ALTER VIEW` statement.

    https://docs.vertica.com/latest/en/sql-reference/statements/alter-statements/alter-view/
    """

    type = "alter_view_statement"

    match_grammar: Matchable = Sequence(
        "ALTER",
        "VIEW",
        Delimited(Ref("TableReferenceSegment")),
        AnyNumberOf(
            Sequence("OWNER", "TO", Ref("ParameterNameSegment")),
            Sequence("SET", "SCHEMA", Ref("SchemaReferenceSegment")),
            Ref("SchemaPrivilegesSegment"),
            Sequence("RENAME", "TO", Delimited(Ref("ParameterNameSegment"))),
        ),
    )


class SetStatementSegment(BaseSegment):
    """Set Statement.

    https://docs.vertica.com/latest/en/sql-reference/statements/set-statements/
    """

    type = "set_statement"

    match_grammar = Sequence(
        "SET",
        OneOf(
            Sequence(
                OneOf(
                    "DATESTYLE",
                    "ESCAPE_STRING_WARNING",
                    "INTERVALSTYLE",
                    "LOCALE",
                    "STANDARD_CONFORMING_STRINGS",
                ),
                "TO",
                Ref("ParameterNameSegment"),
            ),
            Sequence(
                "SEARCH_PATH",
                OneOf("TO", Ref("EqualsSegment")),
                OneOf(Delimited(Ref("ParameterNameSegment")), "DEFAULT"),
            ),
            Sequence(
                "ROLE",
                OneOf(
                    "NONE",
                    "DEFAULT",
                    Sequence(
                        "ALL",
                        Sequence(
                            "EXCEPT",
                            Delimited(Ref("ParameterNameSegment")),
                            optional=True,
                        ),
                    ),
                    Delimited(Ref("ParameterNameSegment")),
                ),
            ),
            Sequence(
                "TIME",
                "ZONE",
                Ref.keyword("TO", optional=True),
                OneOf(Ref("ParameterNameSegment"), Ref("QuotedLiteralSegment")),
            ),
            Sequence(
                Ref.keyword("SESSION", optional=True),
                "RESOURCE_POOL",
                Ref("EqualsSegment"),
                OneOf(
                    Ref("QuotedLiteralSegment"), Ref("ParameterNameSegment"), "DEFAULT"
                ),
            ),
            Sequence(
                "SESSION",
                OneOf(
                    Sequence(
                        "AUTHORIZATION", OneOf(Ref("ParameterNameSegment"), "DEFAULT")
                    ),
                    Sequence("AUTOCOMMIT", "TO", OneOf("ON", "OFF")),
                    Sequence(
                        "CHARACTERISTICS",
                        "AS",
                        "TRANSACTION",
                        Ref("ParameterNameSegment"),
                    ),
                    Sequence(
                        OneOf("GRACEPERIOD", "IDLESESSIONTIMEOUT", "RUNTIMECAP"),
                        OneOf(
                            "NONE",
                            Sequence(Ref("EqualsSegment"), "DEFAULT"),
                            Ref("QuotedLiteralSegment"),
                        ),
                    ),
                    Sequence(
                        OneOf("MEMORYCAP", "TEMPSPACECAP"),
                        OneOf(
                            "NONE",
                            Sequence(Ref("EqualsSegment"), "DEFAULT"),
                            Ref("QuotedLiteralSegment"),
                        ),
                    ),
                    Sequence("MULTIPLEACTIVERESULTSETS", "TO", OneOf("ON", "OFF")),
                    Sequence(
                        "WORKLOAD",
                        Ref.keyword("TO", optional=True),
                        OneOf(Ref("ParameterNameSegment"), "DEFAULT", "NONE"),
                    ),
                ),
            ),
        ),
    )


class CommentOnStatementSegment(BaseSegment):
    """`COMMENT ON` statement.

    https://www.postgresql.org/docs/13/sql-comment.html
    """

    type = "comment_clause"

    match_grammar = Sequence(
        "COMMENT",
        "ON",
        Sequence(
            OneOf(
                Sequence(
                    OneOf(
                        "TABLE",
                        "VIEW",
                        "PROJECTION",
                    ),
                    Ref("TableReferenceSegment"),
                ),
                Sequence(
                    "COLUMN",
                    Ref("ColumnReferenceSegment"),
                ),
                Sequence(
                    "CONSTRAINT",
                    Ref("ObjectReferenceSegment"),
                    Sequence(
                        "ON",
                        Ref("ObjectReferenceSegment"),
                    ),
                ),
                Sequence(
                    OneOf("AGGREGATE", "ANALYTIC", "TRANSFORM", optional=True),
                    "FUNCTION",
                    Ref("FunctionNameSegment"),
                    Sequence(Ref("FunctionParameterListGrammar"), optional=True),
                ),
                Sequence(
                    "SCHEMA",
                    Ref("SchemaReferenceSegment"),
                ),
                Sequence(
                    "NODE",
                    Ref("ParameterNameSegment"),
                ),
                Sequence(
                    OneOf("SEQUENCE", "LIBRARY"),
                    Ref("ObjectReferenceSegment"),
                ),
            ),
            Sequence("IS", OneOf(Ref("QuotedLiteralSegment"), "NULL")),
        ),
    )


class TransactionalStatements(BaseSegment):
    """DML commands wrapped by BEGIN and END.

    As in https://docs.vertica.com/latest/en/sql-reference/statements/begin/
    https://docs.vertica.com/latest/en/sql-reference/statements/end/
    """

    type = "transactional_statement"
    match_grammar: Matchable = Sequence(
        # TODO add rollback, commit logic and optional keywords
        "BEGIN",
        AnyNumberOf(
            Sequence(
                Ref("InsertStatementSegment"),
                Ref("SemicolonSegment"),
            ),
            Sequence(
                Ref("UpdateStatementSegment"),
                Ref("SemicolonSegment"),
            ),
            Sequence(
                Ref("DeleteStatementSegment"),
                Ref("SemicolonSegment"),
            ),
            Sequence(
                Ref("SelectStatementSegment"),
                Ref("SemicolonSegment"),
            ),
        ),
        "END",
    )


class DatatypeSegment(ansi.DatatypeSegment):
    """A data type segment."""

    match_grammar: Matchable = Sequence(
        OneOf(
            # Date / Datetime
            Ref("TimeWithTZGrammar"),
            "DATE",
            "DATETIME",
            "SMALLDATETIME",
            Sequence("INTERVAL", Ref("IntervalLiteralGrammar", optional=True)),
            # Approximate Numeric
            Sequence(
                "DOUBLE",
                "PRECISION",
            ),
            Sequence("FLOAT", Ref("BracketedArguments", optional=True)),
            "FLOAT8",
            "REAL",
            # Exact Numeric
            "INTEGER",
            "INT",
            "BIGINT",
            "INT8",
            "SMALLINT",
            "TINYINT",
            Sequence(
                OneOf("DECIMAL", "NUMERIC", "NUMBER", "MONEY"),
                Ref("BracketedArguments", optional=True),
            ),
            # Spatial
            Sequence(
                OneOf("GEOMETRY", "GEOGRAPHY"),
                Bracketed(Ref("NumericLiteralSegment"), optional=True),
            ),
            # UUID
            "UUID",
            # Text
            Sequence(
                Ref.keyword("LONG", optional=True),
                "VARCHAR",
                Ref("BracketedArguments", optional=True),
            ),
            Sequence("CHAR", Ref("BracketedArguments", optional=True)),
            # Binary types
            OneOf(
                "BINARY",
                Sequence(Ref.keyword("LONG", optional=True), "VARBINARY"),
                "BYTEA",
                "RAW",
            ),
            "BOOLEAN",
            # array types
            OneOf(
                # TODO: need to add an opportunity to specify size of array
                AnyNumberOf(
                    Bracketed(
                        Ref("ExpressionSegment", optional=True), bracket_type="square"
                    )
                ),
                Ref("ArrayTypeSegment"),
                Ref("SizedArrayTypeSegment"),
            ),
            # TODO: add row data type support
            Sequence(
                OneOf(
                    Sequence(
                        OneOf("CHARACTER", "BINARY"),
                        OneOf("VARYING", Sequence("LARGE", "OBJECT")),
                    ),
                    Sequence(
                        # Some dialects allow optional qualification of data types with
                        # schemas
                        Sequence(
                            Ref("SingleIdentifierGrammar"),
                            Ref("DotSegment"),
                            allow_gaps=False,
                            optional=True,
                        ),
                        Ref("DatatypeIdentifierSegment"),
                        allow_gaps=False,
                    ),
                ),
                # There may be no brackets for some data types
                Ref("BracketedArguments", optional=True),
            ),
        ),
    )


class AlterSessionStatements(BaseSegment):
    """An ALTER SESSION statement.

    https://docs.vertica.com/latest/en/sql-reference/statements/alter-statements/alter-session/
    """

    type = "alter_session_statement"
    match_grammar: Matchable = Sequence(
        "ALTER",
        "SESSION",
        OneOf(
            Sequence(
                "SET",
                Ref.keyword("PARAMETER", optional=True),
                OptionallyBracketed(
                    Sequence(
                        Ref("ParameterNameSegment"),
                        Ref("EqualsSegment"),
                        Ref("QuotedLiteralSegment"),
                    ),
                ),
            ),
            Sequence(
                "CLEAR",
                Ref.keyword("PARAMETER", optional=True),
                OneOf(
                    Bracketed(
                        Sequence(
                            Ref("ParameterNameSegment"),
                            Ref("EqualsSegment"),
                            Ref("QuotedLiteralSegment"),
                        ),
                    ),
                    Sequence("PARAMETER", "ALL"),
                ),
            ),
            Sequence(
                "SET",
                "UDPARAMETER",
                Sequence("FOR", Ref("ParameterNameSegment"), optional=True),
                Bracketed(
                    Sequence(
                        Ref("ParameterNameSegment"),
                        Ref("EqualsSegment"),
                        Ref("QuotedLiteralSegment"),
                    ),
                ),
            ),
            Sequence(
                "CLEAR",
                "UDPARAMETER",
                Sequence("FOR", Ref("ParameterNameSegment"), optional=True),
                OneOf(
                    "ALL",
                    Bracketed(Ref("ParameterNameSegment")),
                ),
            ),
        ),
    )


class CopyStatementSegment(BaseSegment):
    """A `COPY` statement.

    As Specified in https://docs.vertica.com/latest/en/sql-reference/statements/copy/
    """

    # TODO: it's not full and requires additional improvements

    type = "copy_statement"

    match_grammar = Sequence(
        "COPY",
        Ref("TableReferenceSegment"),
        OneOf(
            Bracketed(
                Delimited(Ref("CopyColumnOptionsSegment")),
            ),
            Sequence(
                "COLUMN",
                "OPTION",
                Bracketed(
                    Delimited(
                        Sequence(
                            Ref("ColumnReferenceSegment"),
                            Ref("CopyColumnOptionsSegment", optional=True),
                        ),
                    ),
                ),
            ),
            optional=True,
        ),
        "FROM",
        OneOf(
            Sequence(
                Ref.keyword("LOCAL", optional=True),
                "STDIN",
                Ref("CompressionType", optional=True),
                Ref("QuotedLiteralSegment", optional=True),
            ),
            Sequence(
                "LOCAL",
                Ref("QuotedLiteralSegment"),
                Ref("CompressionType", optional=True),
            ),
            Sequence(
                "VERTICA",
                Ref("TableReferenceSegment"),
                Bracketed(Delimited(Ref("ColumnReferenceSegment")), optional=True),
            ),
            Sequence(Delimited(Ref("QuotedLiteralSegment"))),
        ),
        OneOf("NATIVE", Sequence("NATIVE", "VARCHAR"), "ORC", "PARQUET", optional=True),
        Ref("CopyOptionsSegment", optional=True),
    )


class LambdaExpressionSegment(BaseSegment):
    """Lambda function used in a function.

    https://docs.vertica.com/latest/en/sql-reference/functions/array-functions/array-find/
    """

    type = "lambda_function"
    match_grammar = Sequence(
        OneOf(
            Ref("ParameterNameSegment"),
            Bracketed(Delimited(Ref("ParameterNameSegment"))),
        ),
        Ref("LambdaArrowSegment"),
        Ref("ExpressionSegment"),
    )


class FunctionSegment(ansi.FunctionSegment):
    """A scalar or aggregate function.

    https://docs.vertica.com/latest/en/sql-reference/functions/aggregate-functions/
    """

    match_grammar: Matchable = OneOf(
        Sequence(
            # Treat functions which take date parts separately
            # So those functions parse date parts as DatetimeUnitSegment
            # rather than identifiers.
            Sequence(
                Ref("DatePartFunctionNameSegment"),
                Ref("DateTimeFunctionContentsSegment"),
            ),
        ),
        Ref("ColumnsExpressionGrammar"),
        Sequence(
            Sequence(
                Ref(
                    "FunctionNameSegment",
                    exclude=OneOf(
                        Ref("DatePartFunctionNameSegment"),
                        Ref("ColumnsExpressionFunctionNameSegment"),
                        Ref("ValuesClauseSegment"),
                    ),
                ),
                Ref("FunctionContentsSegment"),
            ),
            AnySetOf(Ref("PostFunctionGrammar")),
        ),
    )


class WithinGroupClauseSegment(BaseSegment):
    """A `WITHIN GROUP` clause for some analytic functions.

    https://docs.vertica.com/latest/en/sql-reference/functions/analytic-functions/percentile-cont-analytic/
    """

    type = "within_group_clause_statement"

    match_grammar = Sequence("WITHIN", "GROUP", Bracketed(Ref("OrderByClauseSegment")))


class TimeseriesClauseSegment(BaseSegment):
    """A vertica `TIMESERIES` clause.

    https://docs.vertica.com/latest/en/sql-reference/statements/select/timeseries-clause/
    """

    type = "timeseries_clause_statement"

    match_grammar: Matchable = Sequence(
        "TIMESERIES",
        Ref("AliasExpressionSegment"),
        Ref.keyword("AS"),
        Ref("QuotedLiteralSegment"),
        Indent,
        Ref("OverClauseSegment"),
        # TODO: add optional ORDER BY
        Dedent,
    )


class UnorderedSelectStatementSegment(ansi.UnorderedSelectStatementSegment):
    """A `SELECT` statement without any ORDER clauses or later.

    Copy of ansi class except additional terminator TimeseriesClauseSegment
    """

    match_grammar: Matchable = Sequence(
        Ref("SelectClauseSegment"),
        Ref("FromClauseSegment", optional=True),
        Ref("WhereClauseSegment", optional=True),
        Ref("GroupByClauseSegment", optional=True),
        Ref("HavingClauseSegment", optional=True),
        Ref("OverlapsClauseSegment", optional=True),
        Ref("NamedWindowSegment", optional=True),
        terminators=[
            Ref("SetOperatorSegment"),
            Ref("WithNoSchemaBindingClauseSegment"),
            Ref("WithDataClauseSegment"),
            Ref("OrderByClauseSegment"),
            Ref("LimitClauseSegment"),
            Ref("TimeseriesClauseSegment"),
        ],
        parse_mode=ParseMode.GREEDY_ONCE_STARTED,
    )


class SelectStatementSegment(ansi.SelectStatementSegment):
    """A `SELECT` statement.

    Copy of ansi class except additional TimeseriesClauseSegment grammar
    """

    match_grammar = UnorderedSelectStatementSegment.match_grammar.copy(
        insert=[
            Ref("OrderByClauseSegment", optional=True),
            Ref("FetchClauseSegment", optional=True),
            Ref("LimitClauseSegment", optional=True),
            Ref("TimeseriesClauseSegment", optional=True),
            Ref("NamedWindowSegment", optional=True),
        ],
        # Overwrite the terminators, because we want to remove some.
        replace_terminators=True,
        terminators=[
            Ref("SetOperatorSegment"),
            Ref("WithNoSchemaBindingClauseSegment"),
            Ref("WithDataClauseSegment"),
        ],
    )


class NullEqualsSegment(CompositeComparisonOperatorSegment):
    """Null Equals operator."""

    match_grammar: Matchable = Ref("NullEqualsOperatorSegment")


class PartitionClauseSegment(ansi.PartitionClauseSegment):
    """A `PARTITION BY` for window functions.

    https://docs.vertica.com/latest/en/sql-reference/language-elements/window-clauses/window-partition-clause/
    """

    match_grammar: Matchable = Sequence(
        "PARTITION",
        OneOf(
            Sequence(
                "BY",
                Indent,
                # Brackets are optional in a partition by statement
                OptionallyBracketed(Delimited(Ref("ExpressionSegment"))),
                Dedent,
            ),
            "BEST",
            "NODES",
            "ROW",
            Sequence("LEFT", "JOIN"),
        ),
    )


class FrameClauseSegment(ansi.FrameClauseSegment):
    """A frame clause for window functions.

    https://docs.vertica.com/latest/en/sql-reference/language-elements/window-clauses/window-partition-clause/
    """

    type = "frame_clause"

    _frame_extent = OneOf(
        Sequence("CURRENT", "ROW"),
        Sequence(
            OneOf(
                Ref("NumericLiteralSegment"),
                OneOf(
                    Sequence(
                        Ref("QuotedLiteralSegment"),
                        Ref("CastOperatorSegment"),
                        "INTERVAL",
                    ),
                    Sequence(
                        # TODO maybe this logic should be in an additional segment?
                        # because there are so many options
                        # for the interval representation.
                        Ref.keyword("INTERVAL", optional=True),
                        OneOf(
                            Ref("IntervalLiteralGrammar"), Ref("QuotedLiteralSegment")
                        ),
                        Ref("DatetimeUnitSegment", optional=True),
                    ),
                ),
                "UNBOUNDED",
            ),
            OneOf("PRECEDING", "FOLLOWING"),
        ),
    )

    match_grammar: Matchable = Sequence(
        Ref("FrameClauseUnitGrammar"),
        OneOf(_frame_extent, Sequence("BETWEEN", _frame_extent, "AND", _frame_extent)),
    )


class AlterSchemaStatementSegment(BaseSegment):
    """An `ALTER SCHEMA` statement.

    https://docs.vertica.com/latest/en/sql-reference/statements/alter-statements/alter-schema/
    """

    type = "alter_schema_statement"
    match_grammar = Sequence(
        "ALTER",
        "SCHEMA",
        Delimited(Ref("SchemaReferenceSegment")),
        OneOf(
            Sequence("DEFAULT", Ref("SchemaPrivilegesSegment")),
            Sequence(
                "RENAME",
                "TO",
                Delimited(Ref("SchemaReferenceSegment")),
            ),
            Sequence(
                "OWNER",
                "TO",
                Ref("RoleReferenceSegment"),
                Ref.keyword("CASCADE", optional=True),
            ),
            Ref("DiskQuotaSegment"),
        ),
    )


class CreateSchemaStatementSegment(ansi.CreateSchemaStatementSegment):
    """A `CREATE SCHEMA` statement.

    https://docs.vertica.com/latest/en/sql-reference/statements/create-statements/create-schema/
    """

    match_grammar: Matchable = Sequence(
        "CREATE",
        "SCHEMA",
        Ref("IfNotExistsGrammar", optional=True),
        Ref("SchemaReferenceSegment"),
        AnySetOf(
            Sequence("AUTHORIZATION", Ref("RoleReferenceSegment")),
            Sequence("DEFAULT", Ref("SchemaPrivilegesSegment")),
            Ref("DiskQuotaSegment"),
        ),
    )


class AliasExpressionSegment(ansi.AliasExpressionSegment):
    """A reference to an object with an `AS` clause.

    The optional AS keyword allows both implicit and explicit aliasing.
    """

    match_grammar: Matchable = OneOf(
        Sequence(
            Indent,
            Ref("AsAliasOperatorSegment", optional=True),
            OneOf(
                Sequence(
                    Ref("SingleIdentifierGrammar"),
                    # Column alias in VALUES clause
                    Bracketed(Ref("SingleIdentifierListSegment"), optional=True),
                ),
                Sequence(Bracketed(Ref("SingleIdentifierListSegment"), optional=True)),
                Ref("SingleQuotedIdentifierSegment"),
            ),
            Dedent,
        ),
        # Some functions alias several columns in brackets () like mapkeys or explode
        Sequence(
            Indent,
            Ref("AsAliasOperatorSegment", optional=True),
            Bracketed(Delimited(Ref("ColumnReferenceSegment"))),
            Dedent,
        ),
    )



"""Python wrapper for Rust parser that returns BaseSegment trees.

This module provides RustParser, a drop-in replacement for the Python Parser
that uses the Rust implementation under the hood but returns Python BaseSegment
objects for compatibility with the linter infrastructure.

The Rust parser returns a MatchResult that describes what was matched (slices,
classes, inserts) without building the AST. Python's MatchResult.apply() then
constructs the BaseSegment tree, leveraging proven logic and avoiding
double-counting issues.
"""

import functools
import logging
import os
import time
from collections import defaultdict
from typing import TYPE_CHECKING, Any, Optional, Union

from sqlfluff.core.config import FluffConfig
from sqlfluff.core.errors import SQLParseError
from sqlfluff.core.parser.context import ParseContext
from sqlfluff.core.parser.match_result import MatchResult, _get_point_pos_at_idx
from sqlfluff.core.parser.segments import (
    BaseFileSegment,
    BaseSegment,
    Dedent,
    ImplicitIndent,
    Indent,
    TemplateSegment,
    UnparsableSegment,
)

if TYPE_CHECKING:  # pragma: no cover
    from sqlfluff.core.dialects.base import Dialect

# Instantiate the parser logger
parser_logger = logging.getLogger("sqlfluff.parser")


# --- Sub-stage parse profiling -------------------------------------------------
# RustParser.parse() is a fast Rust core wrapped in O(nodes) Python work. To find
# out where the time actually goes (Rust parse vs. the Python re-conversion vs.
# building the BaseSegment tree vs. building the Rust arena tree (_rs_tree)),
# the four internal stages can be timed individually.
#
# This is OFF by default (zero overhead). Enable it either by setting the
# SQLFLUFF_RS_PROFILE environment variable, or by calling set_profiling(True)
# from tooling (e.g. utils/benchmark_parsing.py). When enabled, the most recent
# parse() call records per-stage wall-clock seconds into _PARSE_PROFILE, readable
# via get_parse_profile().
_PROFILE_ENABLED = bool(os.environ.get("SQLFLUFF_RS_PROFILE"))
_PARSE_PROFILE: dict[str, float] = {}


def set_profiling(enabled: bool) -> None:
    """Enable or disable per-stage parse profiling at runtime."""
    global _PROFILE_ENABLED
    _PROFILE_ENABLED = enabled


def reset_parse_profile() -> None:
    """Clear the accumulated per-stage profile.

    parse() accumulates timings (across rendered variants), so callers reset
    between top-level parses to scope the profile to a single source. The
    benchmark calls this before each timed iteration.
    """
    _PARSE_PROFILE.clear()


def get_parse_profile() -> dict[str, float]:
    """Return per-stage timings (seconds) accumulated since the last reset.

    parse() *accumulates* into this profile (summing across rendered variants,
    e.g. branching Jinja), so it reflects all parse() calls since the last
    reset_parse_profile() rather than just the final one. Only populated when
    profiling is enabled (see set_profiling / the SQLFLUFF_RS_PROFILE env var).
    Keys, in execution order:
        rust_core      - the Rust parse (parse_match_result_from_tokens)
        convert        - Python rebuild of the tree as a MatchResult
        apply          - building the BaseSegment tree (MatchResult.apply)
        apply_as_tree  - building the Rust arena tree (_rs_tree)
    """
    return dict(_PARSE_PROFILE)


# --- Native AST construction (experimental) ------------------------------------
# Today parse() rebuilds the Rust match tree as Python MatchResult objects
# (_convert_rs_match_result) and then walks that again to build the BaseSegment
# tree (MatchResult.apply) - two full traversals plus a throwaway intermediate
# tree. The "native AST" path fuses these into a single pass that instantiates
# BaseSegments directly from the RsMatchResult (_apply_rs_match_result),
# eliminating the intermediate MatchResult tree (the profiler's "convert" stage).
#
# OFF by default while we validate parity against the proven convert+apply path.
# Toggle via the SQLFLUFF_RS_NATIVE_AST env var or set_native_ast(True).
_NATIVE_AST_ENABLED = bool(os.environ.get("SQLFLUFF_RS_NATIVE_AST"))


def set_native_ast(enabled: bool) -> None:
    """Enable or disable the fused (native) BaseSegment builder at runtime."""
    global _NATIVE_AST_ENABLED
    _NATIVE_AST_ENABLED = enabled


def get_native_ast() -> bool:
    """Return whether the fused (native) BaseSegment builder is enabled."""
    return _NATIVE_AST_ENABLED


try:
    from sqlfluffrs import RsMatchResult, RsParseError, RsParser, RsToken

    _HAS_RUST_PARSER = True

    class RustParser:
        """Parser wrapper that uses Rust implementation but returns BaseSegment.

        This is designed to be a drop-in replacement for the Python Parser class,
        with the same interface but using the faster Rust parser internally.

        Example:
            >>> from sqlfluff.core.parser import RustParser
            >>> parser = RustParser(dialect="ansi")
            >>> segment = parser.parse(tokens, fname="test.sql")
        """

        def __init__(
            self,
            config: Optional["FluffConfig"] = None,
            dialect: Optional[str] = None,
        ) -> None:
            """Initialize the Rust parser with config or dialect.

            Args:
                config: FluffConfig instance (takes precedence over dialect)
                dialect: Dialect name (used if config not provided)

            Raises:
                ValueError: If both config and dialect are provided
            """
            if config and dialect:
                raise ValueError(  # pragma: no cover
                    "RustParser does not support setting both `config` and `dialect`."
                )

            # Use the provided config or create one from the dialect.
            self.config = config or FluffConfig.from_kwargs(dialect=dialect)
            self.RootSegment: type[BaseFileSegment] = self.config.get(
                "dialect_obj"
            ).get_root_segment()

            # Extract indentation config and convert boolean values only
            indent_config = self.config.get_section("indentation") or {}
            if indent_config:
                # Only keep boolean config values for conditional evaluation
                # Non-boolean values like "indent_unit": "space" are not needed
                # for conditionals
                indent_config = {
                    k: v for k, v in indent_config.items() if isinstance(v, bool)
                }

            # Max parse depth (DoS mitigation); 0 disables the limit
            max_parse_depth = self.config.get("max_parse_depth")
            max_parse_nodes = self.config.get("max_parse_nodes")
            assert isinstance(max_parse_depth, int)
            assert isinstance(max_parse_nodes, int)
            assert max_parse_depth >= 0
            assert max_parse_nodes >= 0

            # Create the Rust parser
            self._rs_parser = RsParser(
                dialect=self.config.get("dialect"),
                indent_config=indent_config,
                max_parser_iterations=self.config.get("rust_parser_max_iterations")
                or None,
                parser_warn_threshold=self.config.get("rust_parser_warn_threshold")
                or None,
                max_parse_depth=max_parse_depth,
                max_parse_nodes=max_parse_nodes,
            )

        def parse(
            self,
            segments: tuple["BaseSegment", ...],
            fname: Optional[str] = None,
            parse_statistics: bool = False,
        ) -> Optional["BaseSegment"]:
            """Parse a series of lexed tokens using the Rust parser.

            This method now uses the MatchResult-based approach where Rust returns
            a MatchResult and Python's apply() builds the AST, avoiding double-counting.

            Args:
                segments: Tuple of RawSegment objects from the lexer
                fname: Optional filename for error reporting
                parse_statistics: Whether to log parse statistics (not yet implemented)
                tf: Optional TemplatedFile for position marker reconstruction

            Returns:
                BaseSegment tree representing the parsed SQL, or None if empty

            Note:
                The segments parameter contains RawSegment objects that were created
                from RsToken objects using RawSegment.from_rstoken(). To optimize,
                we need to extract the original RsToken objects to pass to the Rust
                parser directly.
            """
            try:
                if not segments:  # pragma: no cover
                    return None

                parse_context = ParseContext.from_config(config=self.config)
                parse_context.seed_parse_nodes(len(segments))

                # Per-stage profiling (no-op unless profiling is enabled).
                # NOTE: parse() runs once per rendered variant, so timings are
                # *accumulated* into _PARSE_PROFILE (see the publish step below)
                # rather than overwritten — branching templates (e.g. Jinja
                # `{% if %}`) parse multiple variants and we want their combined
                # cost, consistent with the linter's total parse time. Callers
                # reset between top-level parses via reset_parse_profile()
                # (the benchmark does this per timed iteration).
                _prof: Optional[dict[str, float]] = {} if _PROFILE_ENABLED else None
                _ts = 0.0

                # PYTHON PARITY: Trim non-code from start (like root_parse)
                _start_idx = 0
                for _start_idx in range(len(segments)):
                    if segments[_start_idx].is_code:
                        break

                # PYTHON PARITY: Trim non-code from end (like root_parse)
                _end_idx = len(segments)
                for _end_idx in range(len(segments), _start_idx - 1, -1):
                    if segments[_end_idx - 1].is_code:
                        break

                if _start_idx == _end_idx:
                    # No code segments - return FileSegment with just non-code
                    parse_context.increment_parse_nodes()
                    return self.RootSegment(segments=segments, fname=fname)

                # Extract the original RsToken objects from the RawSegments
                # PYTHON PARITY: Only parse the code portion (segments[_start_idx:_end_idx])
                # Just like Python's match(segments[:_end_idx], _start_idx, ...)
                tokens = self._extract_tokens_from_segments(
                    segments[_start_idx:_end_idx]
                )

                # Parse using Rust parser to get MatchResult
                # The Rust parser may raise RsParseError for certain parse errors (e.g.,
                # missing closing brackets in terminators). We catch these and convert to
                # SQLParseError. Regular parse errors are embedded in the MatchResult.
                try:
                    if _prof is not None:
                        _ts = time.perf_counter()
                    rs_match = self._rs_parser.parse_match_result_from_tokens(tokens)
                    if _prof is not None:
                        _prof["rust_core"] = time.perf_counter() - _ts
                except RsParseError as e:
                    # Convert Rust parse error to SQLParseError with position info
                    raise SQLParseError.from_rs_parse_error(
                        e, segments[_start_idx:_end_idx]
                    ) from e

                # Build the BaseSegment tree from the Rust match result.
                # PYTHON PARITY: only the code portion (segments[_start_idx:_end_idx])
                # is passed, since match-result indices are relative to it.
                code_segments = segments[_start_idx:_end_idx]
                if _NATIVE_AST_ENABLED:
                    # Fused path: instantiate segments directly from rs_match in a
                    # single pass (no intermediate Python MatchResult tree).
                    if _prof is not None:
                        _ts = time.perf_counter()
                    _matched = self._apply_rs_match_result(
                        rs_match, code_segments, parse_context
                    )
                    if _prof is not None:
                        _prof["apply"] = time.perf_counter() - _ts
                else:
                    # Legacy path: rebuild a Python MatchResult, then apply it.
                    if _prof is not None:
                        _ts = time.perf_counter()
                    match = self._convert_rs_match_result(rs_match, code_segments)
                    if _prof is not None:
                        _prof["convert"] = time.perf_counter() - _ts
                    parser_logger.info("Root Match:\n%s", match)

                    if _prof is not None:
                        _ts = time.perf_counter()
                    _matched = match.apply(code_segments, parse_context=parse_context)
                    if _prof is not None:
                        _prof["apply"] = time.perf_counter() - _ts

                # PYTHON PARITY: Add back any unmatched segments after the match.
                # matched_slice/truthiness are read from rs_match so both build
                # paths agree (mirrors MatchResult.matched_slice / __bool__).
                _m_start, _m_stop = rs_match.matched_slice
                _match_truthy = _m_stop > _m_start or bool(rs_match.insert_segments)
                matched_stop = _start_idx + _m_stop
                _unmatched = segments[matched_stop:_end_idx]

                # PYTHON PARITY: If there are unmatched code segments, wrap them in
                # UnparsableSegment. This matches the logic in FileSegment.root_parse()
                content: tuple[BaseSegment, ...]
                if not _match_truthy:
                    parse_context.increment_parse_nodes()
                    content = (
                        UnparsableSegment(
                            segments[_start_idx:_end_idx],
                            expected=str(self.RootSegment.match_grammar),
                        ),
                    )
                elif _unmatched:
                    _idx = 0
                    for idx, seg in enumerate(_unmatched):
                        if seg.is_code:
                            _idx = idx
                            break
                    else:  # pragma: no cover
                        _idx = len(_unmatched)
                    parse_context.increment_parse_nodes()
                    content = (
                        _matched
                        + _unmatched[:_idx]
                        + (
                            UnparsableSegment(
                                _unmatched[_idx:],
                                expected="Nothing else in FileSegment.",
                            ),
                        )
                    )
                else:
                    content = _matched + _unmatched

                parse_context.increment_parse_nodes()
                result = self.RootSegment(
                    segments[:_start_idx] + content + segments[_end_idx:], fname=fname
                )

                # Build the mutable Rust arena tree (RsTree) from the MatchResult.
                # This is the id-addressable façade tree (RsTree/RsHandle) used by
                # Rust-side linting/fixing to avoid round-tripping through Python's
                # segment tree.
                try:
                    # Extract leading non-code tokens (segments before _start_idx:
                    # whitespace/newlines at the start of the file) so the arena's
                    # flat raw list matches Python's raw_segments ordering exactly.
                    leading_tokens = (
                        self._extract_tokens_from_segments(segments[:_start_idx])
                        if _start_idx
                        else []
                    )
                    # Extract trailing non-code tokens (segments after _end_idx: newline,
                    # end_of_file, etc.) and include them in the arena so that the
                    # reflow/respace rules can correctly detect EOF and trailing newlines.
                    trailing_tokens = (
                        self._extract_tokens_from_segments(segments[_end_idx:])
                        if _end_idx < len(segments)
                        else []
                    )
                    if _prof is not None:
                        _ts = time.perf_counter()
                    result._rs_tree = rs_match.apply_as_tree(
                        tokens,
                        leading=leading_tokens,
                        trailing=trailing_tokens,
                    )
                    if _prof is not None:
                        _prof["apply_as_tree"] = time.perf_counter() - _ts
                except Exception:  # pragma: no cover
                    # Non-critical: if tree building fails, rules fall back to Python
                    parser_logger.warning(
                        f"Unable to apply match result in parse tree for {fname}, falling"
                        " back to Python. Please report this as a bug with the SQL that"
                        " caused it."
                    )
                    result._rs_tree = None

                # Accumulate this variant's per-stage timings into the profile
                # (summing across rendered variants of the same source).
                if _prof is not None:
                    for _stage, _dur in _prof.items():
                        _PARSE_PROFILE[_stage] = _PARSE_PROFILE.get(_stage, 0.0) + _dur

                if parse_statistics:  # pragma: no cover
                    print(
                        "Warning: parse_statistics not yet implemented for Rust parser"
                    )

                return result
            except SQLParseError as err:
                if err.segment is None:
                    anchor = next((seg for seg in segments if seg.is_code), None)
                    if anchor is not None:
                        err = SQLParseError(
                            description=err.description,
                            segment=anchor,
                            ignore=err.ignore,
                            fatal=err.fatal,
                            warning=err.warning,
                        )
                raise err

        @functools.lru_cache(maxsize=128)
        def _get_segment_class_by_name(
            self, segment_name: str
        ) -> Optional[type[BaseSegment]]:
            """Get a segment class by its exact class name from the dialect.

            Args:
                segment_name: The segment CLASS name (e.g., "AsAliasOperatorSegment")

            Returns:
                The segment class, or None if not found or is a grammar

            Note:
                Results are cached per parser instance since dialect is fixed.
            """
            if segment_name == "UnparsableSegment":
                return UnparsableSegment

            # Get the item from the dialect library
            dialect: Dialect = self.config.get("dialect_obj")
            item = dialect.get_segment(segment_name)

            # Verify it's a valid segment class (not a grammar element)
            if (
                item is not None
                and isinstance(item, type)
                and issubclass(item, BaseSegment)
            ):
                return item

            # Not a valid segment class - return None for grammar names
            return None  # pragma: no cover

        def _rs_match_fields(
            self, rs_match: "RsMatchResult"
        ) -> tuple[
            Optional[type["BaseSegment"]],
            dict[str, Any],
            tuple[tuple[int, type], ...],
        ]:
            """Extract (matched_class, segment_kwargs, insert_segments) from a match.

            Shared by both tree-building paths: _convert_rs_match_result (the
            MatchResult-based path) and _apply_rs_match_result (the fused path).
            """
            # Determine matched_class. The Rust parser includes the actual Python
            # class names (from codegen) in matched_class, so we use them directly.
            matched_class: Optional[type["BaseSegment"]] = (
                self._get_segment_class_by_name(rs_match.matched_class)
                if rs_match.matched_class
                else None
            )

            # Build segment_kwargs - optimize by checking first if we need any
            segment_kwargs: dict[str, Any] = {}
            if rs_match.instance_types:
                segment_kwargs["instance_types"] = tuple(rs_match.instance_types)

            # Copy over any segment_kwargs from Rust
            # (e.g., "expected" for UnparsableSegment)
            if rs_match.segment_kwargs:
                segment_kwargs.update(rs_match.segment_kwargs)

            # Set trim_chars from the match result if available
            if rs_match.trim_chars:
                segment_kwargs["trim_chars"] = tuple(rs_match.trim_chars)

            # Set casefold from the match result if available
            if rs_match.casefold:
                if rs_match.casefold == "upper":
                    segment_kwargs["casefold"] = str.upper
                elif rs_match.casefold == "lower":
                    segment_kwargs["casefold"] = str.lower

            # Set quoted_value and escape_replacements for normalization
            if rs_match.quoted_value:  # pragma: no cover
                segment_kwargs["quoted_value"] = rs_match.quoted_value
            if rs_match.escape_replacements:  # pragma: no cover
                segment_kwargs["escape_replacements"] = rs_match.escape_replacements

            # Extract insert_segments (Indent/Dedent meta segments).
            # rs_match.insert_segments contains (idx, seg_type, is_implicit) tuples;
            # these are pre-flattened on the Rust side.
            insert_segments: tuple[tuple[int, type], ...] = ()
            if rs_match.insert_segments:
                insert_segments = tuple(
                    (
                        idx,
                        (
                            ImplicitIndent
                            if (seg_type == "indent" and is_implicit)
                            else (Indent if seg_type == "indent" else Dedent)
                        ),
                    )
                    for idx, seg_type, is_implicit in rs_match.insert_segments
                )

            return matched_class, segment_kwargs, insert_segments

        def _convert_rs_match_result(
            self,
            rs_match: "RsMatchResult",
            segments: tuple["BaseSegment", ...],
            depth: int = 0,
        ) -> MatchResult:
            """Convert Rust MatchResult to Python MatchResult.

            Args:
                rs_match: RsMatchResult from Rust parser
                segments: Segment array for error reporting
                depth: Current recursion depth for debugging

            Returns:
                Python MatchResult with equivalent structure

            """
            start, stop = rs_match.matched_slice
            matched_class, segment_kwargs, insert_segments = self._rs_match_fields(
                rs_match
            )

            # Convert child matches recursively
            # Note: Transparent grammar nodes are now flattened on the Rust side,
            # so we don't need to do it here anymore
            #
            # NOTE: Keep this a plain loop rather than a generator expression:
            # a genexpr would add its own stack frame per nesting level on top
            # of this recursive call, which halves the safe recursion depth.
            # Matches _apply_rs_match_result's plain for loop below, so both
            # AST-building paths tolerate the same nesting depth.
            child_matches_list = []
            for child in rs_match.child_matches:
                child_matches_list.append(
                    self._convert_rs_match_result(child, segments, depth + 1)
                )
            child_matches = tuple(child_matches_list)

            return MatchResult(
                matched_slice=slice(start, stop),
                matched_class=matched_class,
                child_matches=child_matches,
                segment_kwargs=segment_kwargs,
                insert_segments=insert_segments,
            )

        def _apply_rs_match_result(
            self,
            rs_match: "RsMatchResult",
            segments: tuple["BaseSegment", ...],
            parse_context: ParseContext,
        ) -> tuple["BaseSegment", ...]:
            """Build BaseSegments directly from an RsMatchResult in a single pass.

            This fuses _convert_rs_match_result and MatchResult.apply: instead of
            rebuilding the whole match tree as Python MatchResult objects and then
            walking it again, it walks the RsMatchResult once and instantiates the
            BaseSegment tree directly.

            The algorithm mirrors MatchResult.apply() exactly (including the
            parse-node accounting), so it produces an identical tree.
            """
            start, stop = rs_match.matched_slice
            matched_class, segment_kwargs, insert_segments = self._rs_match_fields(
                rs_match
            )

            # NOTE: Accumulate into a list and build the tuple once at the end.
            # `tuple += ...` looks equivalent but isn't: tuples are immutable,
            # so each `+=` allocates a new tuple and copies everything
            # accumulated so far, making a loop of N appends O(N^2). A list
            # amortises appends/extends to O(N), which matters here since this
            # runs at every level of a potentially large tree (e.g. long
            # column/UNION/IN lists produce nodes with many trigger locations).
            result_segments_list: list[BaseSegment] = []

            # Zero-length match: only meta inserts are valid (mirrors apply()).
            # NOTE: Unreachable in practice since the Rust parser stopped
            # emitting zero-length meta child matches (Bracketed now drops
            # direct-child metas for Python parity). Kept as the defensive
            # mirror of apply()'s zero-length branch, including its
            # parse-node accounting.
            if start == stop:  # pragma: no cover
                for idx, seg in insert_segments:
                    assert idx == start, (
                        f"Tried to insert @{idx} outside of matched "
                        f"slice {(start, stop)}"
                    )
                    _pos = _get_point_pos_at_idx(segments, idx)
                    parse_context.increment_parse_nodes()
                    result_segments_list.append(seg(pos_marker=_pos))
                return tuple(result_segments_list)

            # Merge inserts (added first) and child matches into trigger locations.
            trigger_locs: defaultdict[int, list[Union["RsMatchResult", type]]] = (
                defaultdict(list)
            )
            for idx, seg in insert_segments:
                trigger_locs[idx].append(seg)
            for child in rs_match.child_matches:
                trigger_locs[child.matched_slice[0]].append(child)

            max_idx = start
            for idx in sorted(trigger_locs):
                # Include any untouched segments before this trigger.
                if idx > max_idx:
                    result_segments_list.extend(segments[max_idx:idx])
                    max_idx = idx
                elif idx < max_idx:  # pragma: no cover
                    # An outer match contains overlapping child matches: the
                    # Rust parser emitted inconsistent ranges (mirrors apply()).
                    raise ValueError(
                        "Segment skip ahead error. An outer match contains "
                        "overlapping child matches. This MatchResult was "
                        "wrongly constructed."
                    )
                for trigger in trigger_locs[idx]:
                    if isinstance(trigger, type):
                        # A meta segment class (Indent/Dedent/ImplicitIndent).
                        # NOTE: apply() does not count meta inserts here (only in
                        # the zero-length branch above), so neither do we.
                        _pos = _get_point_pos_at_idx(segments, idx)
                        result_segments_list.append(trigger(pos_marker=_pos))
                    else:
                        # A child match - recurse and advance past it.
                        result_segments_list.extend(
                            self._apply_rs_match_result(
                                trigger, segments, parse_context
                            )
                        )
                        max_idx = trigger.matched_slice[1]

            # Anything left after the last trigger.
            if max_idx < stop:
                result_segments_list.extend(segments[max_idx:stop])

            result_segments = tuple(result_segments_list)

            if not matched_class:
                return result_segments

            new_seg = matched_class.from_result_segments(
                result_segments, segment_kwargs
            )
            parse_context.increment_parse_nodes()
            return (new_seg,)

        @staticmethod
        def _template_segment_to_rstoken(segment: TemplateSegment) -> "RsToken":
            """Convert a Python TemplateSegment to a Rust RsToken.

            Creates a template placeholder token with the segment's metadata.
            This allows the Rust parser to handle templated SQL (Jinja, dbt, etc.).

            Args:
                segment: Python TemplateSegment from the lexer

            Returns:
                RsToken representing a template placeholder
            """
            # Extract position marker data
            pm = segment.pos_marker
            source_slice = (pm.source_slice.start, pm.source_slice.stop)
            templated_slice = (pm.templated_slice.start, pm.templated_slice.stop)
            templated_file = pm.templated_file

            # Convert block_uuid to hex string if present
            block_uuid_str = segment.block_uuid.hex if segment.block_uuid else None

            # Create template placeholder token using Rust constructor
            return RsToken.template_placeholder_from_slice(
                source_slice=source_slice,
                templated_slice=templated_slice,
                block_type=segment.block_type,
                _source_str=segment.source_str,  # Passed for API compatibility
                block_uuid=block_uuid_str,
                templated_file=templated_file,
            )

        def _extract_tokens_from_segments(
            self, segments: tuple["BaseSegment", ...]
        ) -> list["RsToken"]:
            """Extract RsToken objects from RawSegments.

            This extracts the cached _rstoken attribute from segments that came from
            the Rust lexer, or converts TemplateSegments to template placeholder tokens.

            Args:
                segments: Tuple of BaseSegment objects

            Returns:
                List of RsToken objects (or compatible token objects)
            """
            tokens = []
            for segment in segments:
                # Check if segment has _rstoken attribute (cached original token)
                if hasattr(segment, "_rstoken"):
                    tokens.append(segment._rstoken)
                elif isinstance(segment, TemplateSegment):
                    # Template segment - create template placeholder token
                    tokens.append(self._template_segment_to_rstoken(segment))
                else:  # pragma: no cover
                    # Cannot reconstruct RsToken from Python segment
                    raise ValueError(
                        f"Cannot extract RsToken from segment {segment!r}. "
                        f"Unsupported segment type: {type(segment).__name__}."
                    )

            return tokens

except ImportError:
    RustParser = None  # type: ignore[assignment, misc]
    _HAS_RUST_PARSER = False



"""Implementation of Rule LT10."""

from typing import Optional

from sqlfluff.core.parser import NewlineSegment, WhitespaceSegment
from sqlfluff.core.rules import BaseRule, LintFix, LintResult, RuleContext
from sqlfluff.core.rules.crawlers import SegmentSeekerCrawler
from sqlfluff.utils.functional import FunctionalContext, sp


class Rule_LT10(BaseRule):
    """``SELECT`` modifiers (e.g. ``DISTINCT``) must be on the same line as ``SELECT``.

    **Anti-pattern**

    .. code-block:: sql

        select
            distinct a,
            b
        from x


    **Best practice**

    .. code-block:: sql

        select distinct
            a,
            b
        from x

    """

    name = "layout.select_modifiers"
    aliases = ("L041",)
    groups = ("all", "core", "layout")
    crawl_behaviour = SegmentSeekerCrawler({"select_clause"})
    is_fix_compatible = True

    def _eval(self, context: RuleContext) -> Optional[LintResult]:
        """Select clause modifiers must appear on same line as SELECT."""
        # We only care about select_clause.
        assert context.segment.is_type("select_clause")

        # Get children of select_clause and the corresponding select keyword.
        child_segments = FunctionalContext(context).segment.children()
        select_keyword = child_segments[0]

        # See if we have a select_clause_modifier.
        select_clause_modifier_seg = child_segments.first(
            sp.is_type("select_clause_modifier")
        )

        # Rule doesn't apply if there's no select clause modifier.
        if not select_clause_modifier_seg:
            return None

        select_clause_modifier = select_clause_modifier_seg[0]

        # Are there any newlines between the select keyword
        # and the select clause modifier.
        leading_newline_segments = child_segments.select(
            select_if=sp.is_type("newline"),
            loop_while=sp.or_(sp.is_whitespace(), sp.is_meta()),
            start_seg=select_keyword,
        )

        # Rule doesn't apply if select clause modifier
        # is already on the same line as the select keyword.
        if not leading_newline_segments:
            return None

        # We should check if there is whitespace before the select clause modifier
        # and remove this during the lint fix.
        leading_whitespace_segments = child_segments.select(
            select_if=sp.is_type("whitespace"),
            loop_while=sp.or_(sp.is_whitespace(), sp.is_meta()),
            start_seg=select_keyword,
        )

        # We should also check if the following select clause element
        # is on the same line as the select clause modifier.
        trailing_newline_segments = child_segments.select(
            select_if=sp.is_type("newline"),
            loop_while=sp.or_(sp.is_whitespace(), sp.is_meta()),
            start_seg=select_clause_modifier,
        )

        # We will insert these segments directly after the select keyword.
        edit_segments = [
            WhitespaceSegment(),
            select_clause_modifier,
        ]
        if not trailing_newline_segments:
            # if the first select clause element is on the same line
            # as the select clause modifier then also insert a newline.
            edit_segments.append(NewlineSegment())

        fixes = []
        # Move select clause modifier after select keyword.
        fixes.append(
            LintFix.create_after(
                anchor_segment=select_keyword,
                edit_segments=edit_segments,
            )
        )

        # Delete original newlines and whitespace between select keyword
        # and select clause modifier.

        # If there is not a newline after the select clause modifier then delete
        # newlines between the select keyword and select clause modifier.
        if not trailing_newline_segments:
            fixes.extend(LintFix.delete(s) for s in leading_newline_segments)
        # If there is a newline after the select clause modifier then delete both the
        # newlines and whitespace between the select keyword and select clause modifier.
        else:
            fixes.extend(
                LintFix.delete(s)
                for s in leading_newline_segments + leading_whitespace_segments
            )
        # Delete the original select clause modifier.
        fixes.append(LintFix.delete(select_clause_modifier))

        # If there is whitespace (on the same line) after the select clause modifier
        # then also delete this.
        trailing_whitespace_segments = child_segments.select(
            select_if=sp.is_whitespace(),
            loop_while=sp.or_(sp.is_type("whitespace"), sp.is_meta()),
            start_seg=select_clause_modifier,
        )
        if trailing_whitespace_segments:
            fixes.extend(LintFix.delete(s) for s in trailing_whitespace_segments)

        return LintResult(
            anchor=context.segment,
            fixes=fixes,
        )



"""A List of MariaDB SQL keywords.

The full list can be queried from MariaDB directly
https://mariadb.com/kb/en/information-schema-keywords-table/

The reserved keywords are listed on
https://mariadb.com/kb/en/reserved-words/
"""

starrocks_reserved_keywords = """
ACCESSIBLE
ADD
ALL
ALTER
ANALYZE
AND
AS
ASC
ASENSITIVE
BEFORE
BETWEEN
BIGINT
BINARY
BLOB
BOTH
BY
CALL
CASCADE
CASE
CHANGE
CHANGED_PAGE_BITMAPS
CHAR
CHARACTER
CHECK
CLIENT_STATISTICS
COLLATE
COLUMN
CONDITION
CONSTRAINT
CONTINUE
CONVERT
CREATE
CROSS
CURRENT_DATE
CURRENT_ROLE
CURRENT_TIME
CURRENT_TIMESTAMP
CURRENT_USER
CURSOR
DATABASE
DATABASES
DATE_TRUNC
DAY_HOUR
DAY_MICROSECOND
DAY_MINUTE
DAY_SECOND
DEC
DECIMAL
DECLARE
DEFAULT
DELAYED
DELETE
DELETE_DOMAIN_ID
DESC
DESCRIBE
DETERMINISTIC
DISTINCT
DISTINCTROW
DIV
DO_DOMAIN_IDS
DOUBLE
DROP
DUAL
EACH
ELSE
ELSEIF
ENCLOSED
ESCAPED
EXCEPT
EXISTS
EXIT
EXPLAIN
FALSE
FETCH
FLOAT
FLOAT4
FLOAT8
FOR
FORCE
FOREIGN
FROM
FULLTEXT
GENERAL
GRANT
GROUP
HAVING
HIGH_PRIORITY
HOUR_MICROSECOND
HOUR_MINUTE
HOUR_SECOND
IF
IGNORE
IGNORE_DOMAIN_IDS
IGNORE_SERVER_IDS
IN
INDEX
INDEX_STATISTICS
INFILE
INNER
INOUT
INSENSITIVE
INSERT
INT
INT1
INT2
INT3
INT4
INT8
INTEGER
INTERSECT
INTERVAL
INTO
IS
ITERATE
JOIN
KEY
KEYS
KILL
LEADING
LEAVE
LEFT
LIKE
LIMIT
LINEAR
LINES
LOAD
LOCALTIME
LOCALTIMESTAMP
LOCK
LONG
LONGBLOB
LONGTEXT
LOOP
LOW_PRIORITY
MASTER_HEARTBEAT_PERIOD
MASTER_SSL_VERIFY_SERVER_CERT
MATCH
MAXVALUE
MEDIUMBLOB
MEDIUMINT
MEDIUMTEXT
MIDDLEINT
MINUTE_MICROSECOND
MINUTE_SECOND
MOD
MODIFIES
NATURAL
NOT
NO_WRITE_TO_BINLOG
NULL
NUMERIC
OFFSET
OLAP
ON
OPTIMIZE
OPTION
OPTIONALLY
OR
ORDER
OUT
OUTER
OUTFILE
OVER
PAGE_CHECKSUM
PARSE_VCOL_EXPR
PARTITION
PRECISION
PRIMARY
PROCEDURE
PROPERTIES
PURGE
RANGE
READ
READS
READ_WRITE
REAL
RECURSIVE
REF_SYSTEM_ID
REFERENCES
REGEXP
RELEASE
RENAME
REPEAT
REPLACE
REQUIRE
RESIGNAL
RESTRICT
RETURN
RETURNING
REVOKE
RIGHT
RLIKE
ROW_NUMBER
ROWS
SCHEMA
SCHEMAS
SECOND_MICROSECOND
SELECT
SENSITIVE
SEPARATOR
SET
SHOW
SIGNAL
SLOW
SMALLINT
SPATIAL
SPECIFIC
SQL
SQLEXCEPTION
SQLSTATE
SQLWARNING
SQL_BIG_RESULT
SQL_CALC_FOUND_ROWS
SQL_SMALL_RESULT
SSL
STARTING
STATS_AUTO_RECALC
STATS_PERSISTENT
STATS_SAMPLE_PAGES
STRAIGHT_JOIN
TABLE
TERMINATED
THEN
TIME_SLICE
TINYBLOB
TINYINT
TINYTEXT
TO
TRAILING
TRIGGER
TRUE
UNDO
UNION
UNIQUE
UNLOCK
UNSIGNED
UPDATE
USAGE
USE
USING
UTC_DATE
UTC_TIME
UTC_TIMESTAMP
VALUES
VARBINARY
VARCHAR
VARCHARACTER
VARYING
WHEN
WHERE
WHILE
WINDOW
WITH
WRITE
XOR
YEAR_MONTH
ZEROFILL
"""

starrocks_unreserved_keywords = """
ACCOUNT
ACTION
ADMIN
AFTER
AGAINST
AGGREGATE
ALGORITHM
ALWAYS
ANY
ASCII
AT
ATOMIC
AUTHORS
AUTO
AUTO_INCREMENT
AUTOEXTEND_SIZE
AVG
AVG_ROW_LENGTH
BACKUP
BEGIN
BINLOG
BIT
BLOCK
BODY
BOOL
BOOLEAN
BTREE
BROKER
BYTE
CACHE
CASCADED
CATALOG_NAME
CHAIN
CHANGED
CHANNEL
CHARSET
CHECKPOINT
CHECKSUM
CIPHER
CLASS_ORIGIN
CLIENT
CLOB
CLOSE
COALESCE
CODE
COLLATION
COLUMN_NAME
COLUMNS
COLUMN_ADD
COLUMN_CHECK
COLUMN_CREATE
COLUMN_DELETE
COLUMN_GET
COMMENT
COMMIT
COMMITTED
COMPACT
COMPLETION
COMPRESSED
CONCURRENT
CONNECTION
CONSISTENT
CONSTRAINT_CATALOG
CONSTRAINT_NAME
CONSTRAINT_SCHEMA
CONTAINS
CONTEXT
CONTRIBUTORS
CPU
CUBE
CURRENT
CURRENT_POS
CURSOR_NAME
CYCLE
DATA
DATAFILE
DATE
DATETIME
DAY
DEALLOCATE
DEFINER
DELAY_KEY_WRITE
DES_KEY_FILE
DIAGNOSTICS
DIRECTORY
DISABLE
DISCARD
DISK
DO
DUMPFILE
DUPLICATE
DYNAMIC
ELSIF
EMPTY
ENABLE
END
ENDS
ENGINE
ENGINES
ENUM
ERROR
ERRORS
ESCAPE
EVENT
EVENTS
EVERY
EXAMINED
EXCHANGE
EXCLUDE
EXECUTE
EXCEPTION
EXPANSION
EXPIRE
EXPORT
EXTENDED
EXTENT_SIZE
FAST
FAULTS
FEDERATED
FIELDS
FILE
FIRST
FIXED
FLUSH
FOLLOWING
FOLLOWS
FORMAT
FOUND
FULL
FUNCTION
GENERATED
GET_FORMAT
GET
GLOBAL
GOTO
GRANTS
HANDLER
HARD
HASH
HELP
HISTORY
HOST
HOSTS
HOUR
ID
IDENTIFIED
IGNORED
IMMEDIATE
IMPORT
INCREMENT
INDEXES
INITIAL_SIZE
INSERT_METHOD
INSTALL
INVISIBLE
IO
IO_THREAD
IPC
ISOLATION
ISOPEN
ISSUER
INVOKER
JSON
JSON_TABLE
KAFKA
LABEL
KEY_BLOCK_SIZE
LANGUAGE
LAST
LAST_VALUE
LASTVAL
LEAVES
LESS
LEVEL
LIST
LOCAL
LOCKED
LOCKS
LOGFILE
LOGS
MASTER
MASTER_CONNECT_RETRY
MASTER_DELAY
MASTER_GTID_POS
MASTER_HOST
MASTER_LOG_FILE
MASTER_LOG_POS
MASTER_PASSWORD
MASTER_PORT
MASTER_SERVER_ID
MASTER_SSL
MASTER_SSL_CA
MASTER_SSL_CAPATH
MASTER_SSL_CERT
MASTER_SSL_CIPHER
MASTER_SSL_CRL
MASTER_SSL_CRLPATH
MASTER_SSL_KEY
MASTER_USER
MASTER_USE_GTID
MASTER_DEMOTE_TO_SLAVE
MAX_CONNECTIONS_PER_HOUR
MAX_QUERIES_PER_HOUR
MAX_ROWS
MAX_SIZE
MAX_STATEMENT_TIME
MAX_UPDATES_PER_HOUR
MAX_USER_CONNECTIONS
MEDIUM
MEMORY
MERGE
MESSAGE_TEXT
MICROSECOND
MIGRATE
MINUS
MINUTE
MINVALUE
MIN_ROWS
MODE
MODIFY
MONITOR
MONTH
MUTEX
MYSQL
MYSQL_ERRNO
NAME
NAMES
NATIONAL
NCHAR
NESTED
NEVER
NEXT
NEXTVAL
NO
NOMAXVALUE
NOMINVALUE
NOCACHE
NOCYCLE
NO_WAIT
NOWAIT
NODEGROUP
NONE
NOTFOUND
NUMBER
NVARCHAR
OF
OLD_PASSWORD
ONE
ONLINE
ONLY
OPEN
OPTIONS
ORDINALITY
OTHERS
OVERLAPS
OVERWRITE
OWNER
PACKAGE
PACK_KEYS
PAGE
PARSER
PATH
PERIOD
PAUSE
PARTIAL
PARTITIONING
PARTITIONS
PASSWORD
PERSISTENT
PHASE
PLUGIN
PLUGINS
PORT
PORTION
PRECEDES
PRECEDING
PREPARE
PRESERVE
PREV
PREVIOUS
PRIVILEGES
PROCESS
PROCESSLIST
PROFILE
PROFILES
PROXY
QUARTER
QUERY
QUERY_RESPONSE_TIME
QUICK
RAISE
RAW
READ_ONLY
REBUILD
RECOVER
REDO_BUFFER_SIZE
REDOFILE
REDUNDANT
RELAY
RELAYLOG
RELAY_LOG_FILE
RELAY_LOG_POS
RELAY_THREAD
RELOAD
REMOVE
REORGANIZE
REPAIR
REPEATABLE
REPLAY
REPLICA
REPLICAS
REPLICA_POS
REPLICATION
RESET
RESTART
RESTORE
RESUME
RETURNED_SQLSTATE
RETURNS
REUSE
REVERSE
ROLE
ROLLBACK
ROLLUP
ROUTINE
ROW
ROWCOUNT
ROWNUM
ROWTYPE
ROW_COUNT
ROW_FORMAT
RTREE
SAVEPOINT
SCHEDULE
SCHEMA_NAME
SECOND
SECURITY
SEQUENCE
SERIAL
SERIALIZABLE
SESSION
SERVER
SETVAL
SHARE
SHUTDOWN
SIGNED
SIMPLE
SKIP
SLAVE
SLAVES
SLAVE_POS
SNAPSHOT
SOCKET
SOFT
SOME
SONAME
SOUNDS
SOURCE
STAGE
STORED
SQL_AFTER_GTIDS
SQL_BEFORE_GTIDS
SQL_BUFFER_RESULT
SQL_CACHE
SQL_NO_CACHE
SQL_THREAD
SQL_TSI_SECOND
SQL_TSI_MINUTE
SQL_TSI_HOUR
SQL_TSI_DAY
SQL_TSI_WEEK
SQL_TSI_MONTH
SQL_TSI_QUARTER
SQL_TSI_YEAR
START
STARTS
STATEMENT
STATUS
STOP
STORAGE
STRING
SUBCLASS_ORIGIN
SUBJECT
SUBPARTITION
SUBPARTITIONS
SUPER
SUSPEND
SWAPS
SWITCHES
SYSDATE
SYSTEM
SYSTEM_TIME
TABLE_NAME
TABLES
TABLESPACE
TABLE_CHECKSUM
TABLE_STATISTICS
TEMPORARY
TEMPTABLE
TEXT
THAN
TIES
TIME
TIMESTAMP
TIMESTAMPADD
TIMESTAMPDIFF
TRANSACTION
TRANSACTIONAL
THREADS
TRIGGERS
TRUNCATE
TYPE
UNBOUNDED
UNCOMMITTED
UNDEFINED
UNDO_BUFFER_SIZE
UNDOFILE
UNICODE
UNKNOWN
UNINSTALL
UNTIL
UPGRADE
USER
USER_RESOURCES
USER_STATISTICS
USER_VARIABLES
USE_FRM
VALIDATION
VALUE
VARCHAR2
VARIABLES
VIA
VIEW
VIRTUAL
VISIBLE
VERSIONING
WAIT
WARNINGS
WEEK
WEIGHT_STRING
WITHIN
WITHOUT
WORK
WRAPPER
X509
XA
XML
YEAR
"""

# These are not MariaDB keywords, but needed to parse well.
# Taken from the mysql dialect
starrocks_unreserved_keywords += """
NOW
SHARED
INPLACE
NOCOPY
INSTANT
"""

# StarRocks complex/semi-structured type keywords
starrocks_unreserved_keywords += """
ARRAY
MAP
STRUCT
"""



"""Implementation of Rule CV01."""

from typing import Optional

from sqlfluff.core.parser import SymbolSegment
from sqlfluff.core.rules import BaseRule, LintFix, LintResult, RuleContext
from sqlfluff.core.rules.crawlers import SegmentSeekerCrawler
from sqlfluff.utils.functional import FunctionalContext, sp


class Rule_CV01(BaseRule):
    """Consistent usage of ``!=`` or ``<>`` for "not equal to" operator.

    **Anti-pattern**

    .. code-block:: sql

        SELECT * FROM X WHERE 1 <> 2 AND 3 != 4;

    **Best practice**

    Ensure all "not equal to" comparisons are consistent, not mixing ``!=`` and ``<>``.

    .. code-block:: sql

        SELECT * FROM X WHERE 1 != 2 AND 3 != 4;

    """

    name = "convention.not_equal"
    aliases = ("L061",)
    groups = ("all", "convention")
    crawl_behaviour = SegmentSeekerCrawler({"comparison_operator"})
    config_keywords = ["preferred_not_equal_style"]
    is_fix_compatible = True

    def _eval(self, context: RuleContext) -> Optional[LintResult]:
        """Enforce consistent "not equal to" style."""
        self.preferred_not_equal_style: str

        # Rule disabled for teradata as != is not supported.
        if context.dialect.name == "teradata":
            return None

        # Get the comparison operator children
        raw_comparison_operators = (
            FunctionalContext(context)
            .segment.children()
            .select(select_if=sp.is_type("raw_comparison_operator"))
        )

        # Only check ``<>`` or ``!=`` operators
        raw_operator_list = [r.raw for r in raw_comparison_operators]
        if raw_operator_list not in [["<", ">"], ["!", "="]]:
            return None

        memory = context.memory
        # If style is consistent, add the style of the first occurrence to memory
        if self.preferred_not_equal_style == "consistent":
            preferred_not_equal_style = context.memory.get("preferred_not_equal_style")
            if not preferred_not_equal_style:
                preferred_not_equal_style = (
                    "ansi" if raw_operator_list == ["<", ">"] else "c_style"
                )
                memory["preferred_not_equal_style"] = preferred_not_equal_style
        else:
            preferred_not_equal_style = self.preferred_not_equal_style

        if preferred_not_equal_style == "c_style":
            replacement = ["!", "="]
        elif preferred_not_equal_style == "ansi":
            replacement = ["<", ">"]

        # This operator already matches the existing style
        if raw_operator_list == replacement:
            return LintResult(memory=memory)

        # Provide a fix and replace ``<>`` with ``!=``
        # As each symbol is a separate symbol this is done in two steps:
        # Depending on style type, flip any inconsistent operators
        # 1. Flip < and !
        # 2. Flip > and =
        fixes = [
            LintFix.replace(
                raw_comparison_operators[0],
                [SymbolSegment(raw=replacement[0], type="raw_comparison_operator")],
            ),
            LintFix.replace(
                raw_comparison_operators[1],
                [SymbolSegment(raw=replacement[1], type="raw_comparison_operator")],
            ),
        ]

        return LintResult(anchor=context.segment, fixes=fixes, memory=memory)



"""A list of all SQL keywords.

Based on query:
select *
from keywords
"""

# From reserved keywords were removed keywords that work correctly
# in create table(keyword some_datatype) without quotes.
# Put them in unreserved_keywords.

vertica_reserved_keywords = """ALL
AND
ANY
ARRAY
AS
ASC
AUTHORIZATION
BETWEEN
BINARY
BOTH
CASE
CAST
CHECK
COLLATE
COLUMN
CONSTRAINT
CORRELATION
CREATE
CURRENT_DATABASE
CURRENT_DATE
CURRENT_SCHEMA
CURRENT_TIME
CURRENT_TIMESTAMP
CURRENT_USER
DEFAULT
DEFERRABLE
DESC
DISTINCT
ELSE
END
EXCEPT
FALSE
FOR
FOREIGN
FROM
GRANT
HAVING
IN
INITIALLY
INTERSECT
INTERVAL
INTERVALYM
INTO
IS
LATERAL
LEADING
LIKE
LIMIT
LOCALTIME
LOCALTIMESTAMP
MATCH
NEW
NOT
NOTNULL
NULL
NULLSEQUAL
OFFSET
OLD
ON
ONLY
OR
ORDER
OVER
OVERLAPS
PRIMARY
REFERENCES
SELECT
SESSION_USER
SIMILAR
SOME
SYSDATE
TABLE
THEN
TIMESERIES
TO
TRAILING
TRUE
UNBOUNDED
UNION
UNIQUE
USER
USING
WHEN
WHERE
WINDOW
WITH
WITHIN
"""

vertica_unreserved_keywords = """ABORT
ABSOLUTE
ACCESS
ACCESSRANK
ACCOUNT
ACTION
ACTIVATE
ACTIVEPARTITIONCOUNT
ADD
ADDRESS
ADMIN
AFTER
AGGREGATE
ALSO
ALTER
ANALYSE
ANALYTIC
ANALYZE
ANNOTATED
ANTI
ARCHIVE
ASSERTION
ASSIGNMENT
AT
AUTHENTICATION
AUTO
AUTOCOMMIT
AUTO_INCREMENT
AVAILABLE
BACKWARD
BALANCE
BASENAME
BATCH
BEFORE
BEGIN
BEST
BIGINT
BIT
BLOCK
BLOCKDICT_COMP
BLOCK_DICT
BOOLEAN
BROADCAST
BUNDLE
BY
BYTEA
BYTES
BZIP
BZIP_COMP
CA
CACHE
CALL
CALLED
CASCADE
CATALOGPATH
CERTIFICATE
CERTIFICATES
CHAIN
CHAR
CHARACTER
CHARACTERISTICS
CHARACTERS
CHARACTER_LENGTH
CHAR_LENGTH
CHECKPOINT
CIPHER
CLASS
CLEAR
CLOSE
CLUSTER
COLLECTIONCLOSE
COLLECTIONDELIMITER
COLLECTIONENCLOSE
COLLECTIONNULLELEMENT
COLLECTIONOPEN
COLSIZES
COLUMNS
COLUMNS_COUNT
COMMENT
COMMIT
COMMITTED
COMMONDELTA_COMP
COMMUNAL
COMPLEX
CONFIGURATION
CONNECT
CONSTRAINTS
CONTROL
COPY
COUNT
CPUAFFINITYMODE
CPUAFFINITYSET
CREATEDB
CREATEUSER
CRON
CROSS
CSV
CUBE
CURRENT
CURSOR
CUSTOM
CUSTOM_PARTITIONS
CYCLE
DATA
DATABASE
DATAPATH
DATEDIFF
DATESTYLE
DATETIME
DATETIMES
DAY
DEACTIVATE
DEALLOCATE
DEBUG
DEC
DECIMAL
DECLARE
DECODE
DEFAULTS
DEFERRED
DEFINE
DEFINER
DELETE
DELIMITED
DELIMITER
DELIMITERS
DELTARANGE_COMP
DELTARANGE_COMP_SP
DELTAVAL
DEPENDS
DETERMINES
DIRECT
DIRECTCOLS
DIRECTED
DIRECTGROUPED
DIRECTPROJ
DISABLE
DISABLED
DISCONNECT
DISK_QUOTA
DISTVALINDEX
DO
DOMAIN
DOUBLE
DROP
DRYRUN
DURABLE
EACH
ENABLE
ENABLED
ENCLOSED
ENCODED
ENCODING
ENCRYPTED
ENFORCELENGTH
EPHEMERAL
EPOCH
ERROR
ESCAPE
ESCAPE_STRING_WARNING
EVENT
EVENTS
EXCEPTION
EXCEPTIONS
EXCLUDE
EXCLUDING
EXCLUSIVE
EXECUTE
EXECUTIONPARALLELISM
EXISTS
EXPIRE
EXPLAIN
EXPORT
EXTEND
EXTENSIONS
EXTERNAL
EXTRACT
FAILED_LOGIN_ATTEMPTS
FALLTHROUGH
FAULT
FENCED
FETCH
FILESYSTEM
FILLER
FILTER
FIRST
FIXEDWIDTH
FLEX
FLEXIBLE
FLOAT
FOLLOWING
FORCE
FORMAT
FORWARD
FREEZE
FULL
FUNCTION
FUNCTIONS
GCDDELTA
GEOMETRY
GEOGRAPHY
GET
GLOB
GLOBAL
GRACEPERIOD
GROUP
GROUPED
GROUPING
GZIP
GZIP_COMP
HANDLER
HCATALOG
HCATALOG_CONNECTION_TIMEOUT
HCATALOG_DB
HCATALOG_SCHEMA
HCATALOG_SLOW_TRANSFER_LIMIT
HCATALOG_SLOW_TRANSFER_TIME
HCATALOG_USER
HIGH
HIVESERVER2_HOSTNAME
HOLD
HOST
HOSTNAME
HOUR
HOURS
ICEBERG
ID
IDENTIFIED
IDENTITY
IDLESESSIONTIMEOUT
IF
IGNORE
ILIKE
ILIKEB
IMMEDIATE
IMMUTABLE
IMPLICIT
INCLUDE
INCLUDING
INCREMENT
INDEX
INHERITS
INNER
INOUT
INPUT
INSENSITIVE
INSERT
INSTEAD
INT
INTEGER
INTERFACE
INTERPOLATE
INTERVALSTYLE
INVOKER
ISNULL
ISOLATION
JOIN
JSON
KEY
KEYMANAGER
KSAFE
LABEL
LANCOMPILER
LANGUAGE
LARGE
LAST
LATEST
LEFT
LENGTH
LESS
LEVEL
LIBRARY
LIKEB
LISTEN
LOAD
LOADER
LOCAL
LOCALE
LOCATION
LOCK
LONG
LOW
LZO
MANAGED
MAP
MASK
MATCHED
MATERIALIZE
MAXCONCURRENCY
MAXCONCURRENCYGRACE
MAXCONNECTIONS
MAXMEMORYSIZE
MAXPAYLOAD
MAXQUERYMEMORYSIZE
MAXVALUE
MEDIUM
MEMORYCAP
MEMORYSIZE
MERGE
MERGEOUT
METHOD
MICROSECONDS
MILLISECONDS
MINUS
MINUTE
MINUTES
MINVALUE
MODE
MODEL
MONEY
MONTH
MOVE
MOVEOUT
MULTIPLEACTIVERESULTSETS
NAME
NAMESPACE
NATIONAL
NATIVE
NATURAL
NCHAR
NETWORK
NEXT
NO
NOCREATEDB
NOCREATEUSER
NODE
NODES
NOGLOB
NONE
NOTHING
NOTIFIER
NOTIFY
NOWAIT
NULLAWARE
NULLCOLS
NULLS
NUMBER
NUMERIC
OBJECT
OBJECTS
OCTETS
OF
OFF
OIDS
OPERATOR
OPT
OPTIMIZER
OPTION
OPTVER
ORC
OTHERS
OUT
OUTER
OVERLAY
OWNER
PARAMETER
PARAMETERS
PARQUET
PARSER
PARTIAL
PARTITION
PARTITIONING
PASSWORD
PASSWORD_GRACE_TIME
PASSWORD_LIFE_TIME
PASSWORD_LOCK_TIME
PASSWORD_MAX_LENGTH
PASSWORD_MIN_CHAR_CHANGE
PASSWORD_MIN_DIGITS
PASSWORD_MIN_LENGTH
PASSWORD_MIN_LETTERS
PASSWORD_MIN_LIFE_TIME
PASSWORD_MIN_LOWERCASE_LETTERS
PASSWORD_MIN_SYMBOLS
PASSWORD_MIN_UPPERCASE_LETTERS
PASSWORD_REUSE_MAX
PASSWORD_REUSE_TIME
PATTERN
PERCENT
PERMANENT
PINNED
PLACING
PLANNEDCONCURRENCY
POINT
POLICY
POOL
PORT
POSITION
PRECEDING
PRECISION
PREFER
PREFIX
PREPARE
PREPASS
PRESERVE
PREVIOUS
PRIOR
PRIORITY
PRIVILEGES
PROCEDURAL
PROCEDURE
PROFILE
PROJECTION
PROJECTIONS
PSDATE
QUERY
QUEUETIMEOUT
QUOTE
RANDOM
RANGE
RAW
READ
REAL
RECHECK
RECORD
RECOVER
RECURSIVE
REFRESH
REINDEX
REJECTED
REJECTMAX
RELATIVE
RELEASE
REMOVE
RENAME
REORGANIZE
REPEATABLE
REPLACE
REPLICATE
RESET
RESOURCE
RESOURCE_POOL
RESTART
RESTORE
RESTRICT
RESULTS
RETENTION
RETRY
RETURN
RETURNREJECTED
REVOKE
RIGHT
RLE
ROLE
ROLES
ROLLBACK
ROLLUP
ROUTE
ROUTING
ROW
ROWS
RULE
RUNTIMECAP
RUNTIMEPRIORITY
RUNTIMEPRIORITYTHRESHOLD
RWITH
SALT
SAVE
SAVEPOINT
SCHEDULE
SCHEMA
SCHEMATA
SCROLL
SEARCH_PATH
SECOND
SECONDARY
SECONDS
SECURITY
SECURITY_ALGORITHM
SEGMENTED
SEMI
SEMIALL
SEQUENCE
SEQUENCES
SERIAL
SERIALIZABLE
SESSION
SET
SETOF
SETS
SHARD
SHARE
SHARED
SHOW
SIGNED
SIMPLE
SINGLEINITIATOR
SITE
SITES
SKIP
SMALLDATETIME
SMALLINT
SOURCE
SPLIT
SSL_CONFIG
STABLE
STANDARD_CONFORMING_STRINGS
STANDBY
START
STATEMENT
STATISTICS
STDIN
STDOUT
STEMMER
STORAGE
STORED
STREAM
STRENGTH
STRICT
SUBCLUSTER
SUBJECT
SUBNET
SUBSTRING
SUITES
SYSID
SYSTEM
TABLES
TABLESAMPLE
TABLESPACE
TEMP
TEMPLATE
TEMPORARY
TEMPSPACECAP
TERMINATOR
TEXT
THAN
TIES
TIME
TIMESTAMP
TIMESTAMPADD
TIMESTAMPDIFF
TIMESTAMPTZ
TIMETZ
TIMEZONE
TINYINT
TLS
TLSMODE
TOAST
TOKENIZER
TOLERANCE
TRANSACTION
TRANSFORM
TREAT
TRICKLE
TRIGGER
TRIM
TRUNCATE
TRUSTED
TUNING
TYPE
TYPES
UDPARAMETER
UNCOMMITTED
UNCOMPRESSED
UNI
UNINDEXED
UNKNOWN
UNLIMITED
UNLISTEN
UNLOCK
UNSEGMENTED
UPDATE
USAGE
UUID
VACUUM
VALID
VALIDATE
VALIDATOR
VALINDEX
VALUE
VALUES
VARBINARY
VARCHAR
VARCHAR2
VARYING
VERBOSE
VERTICA
VIEW
VOLATILE
WAIT
WEBHDFS_ADDRESS
WEBSERVICE_HOSTNAME
WEBSERVICE_PORT
WITHOUT
WORK
WORKLOAD
WRITE
YEAR
ZONE
ZSTD
ZSTD_COMP
ZSTD_FAST_COMP
ZSTD_HIGH_COMP
"""



"""Methods for loading config files with an ini-style format.

This includes `.sqlfluff` and `tox.ini` files.
"""

import configparser

from sqlfluff.core.helpers.dict import NestedDictRecord, records_to_nested_dict
from sqlfluff.core.types import ConfigMappingType, ConfigValueType


def coerce_value(val: str) -> ConfigValueType:
    """Try to coerce to a more specific type."""
    # Try to coerce it to a more specific type,
    # otherwise just make it a string.
    v: ConfigValueType
    try:
        v = int(val)
    except ValueError:
        try:
            v = float(val)
        except ValueError:
            cleaned_val = val.strip().lower()
            if cleaned_val == "true":
                v = True
            elif cleaned_val == "false":
                v = False
            elif cleaned_val == "none":
                v = None
            else:
                v = val
    return v


def load_ini_string(cfg_content: str) -> ConfigMappingType:
    """Read an ini-style config string.

    This would include loading a `.sqlfluff` file.

    Notes:
    - We rename the root `sqlfluff` section, to `core` so that it's in
      line with other config files.
    - The `configparser` reads everything as strings, but this method will
      attempt to find better types for values based on their content.
    - Path resolution isn't done here, that all happens later.
    - Unlike most cfg file readers, SQLFluff is case-sensitive in how
      it reads config files. This is to ensure we support the case
      sensitivity of jinja.
    """
    # If the string is empty, no need to parse it.
    if not cfg_content:
        return {}

    # Disable interpolation so we can load macros
    config = configparser.ConfigParser(delimiters="=", interpolation=None)
    # NB: We want to be case sensitive in how we read from files,
    # because jinja is also case sensitive. To do this we override
    # the optionxform attribute.
    config.optionxform = lambda option: option  # type: ignore

    # Read the content.
    config.read_string(cfg_content)

    # Build up a buffer of config values.
    config_buffer: list[NestedDictRecord[ConfigValueType]] = []
    for k in config.sections():
        if k == "sqlfluff":
            key: tuple[str, ...] = ("core",)
        elif k.startswith("sqlfluff:"):
            # Return a tuple of nested values
            key = tuple(k[len("sqlfluff:") :].split(":"))
        else:  # pragma: no cover
            # if it doesn't start with sqlfluff, then ignore this
            # section. It's not relevant to sqlfluff.
            continue

        for name, val in config.items(section=k):
            # Try to coerce it to a more specific type,
            # otherwise just make it a string.
            v = coerce_value(val)

            # Split the name on dots to support nested structures.
            # For example, 'namespace.projectname' becomes ('namespace', 'projectname')
            # This allows config like:
            #   [sqlfluff:templater:jinja:context]
            #   namespace.projectname=test
            # to create the nested structure: {'namespace': {'projectname': 'test'}}
            name_parts = tuple(name.split("."))

            # Add the name parts to the end of the key
            config_buffer.append((key + name_parts, v))

    # Compress that buffer into a dictionary.
    return records_to_nested_dict(config_buffer)



"""Implementation of Rule ST06."""

from collections.abc import Iterator
from typing import Optional, Union

from sqlfluff.core.parser import BaseSegment
from sqlfluff.core.rules import (
    BaseRule,
    EvalResultType,
    LintFix,
    LintResult,
    RuleContext,
)
from sqlfluff.core.rules.crawlers import SegmentSeekerCrawler


class Rule_ST06(BaseRule):
    """Select wildcards then simple targets before calculations and aggregates.

    **Anti-pattern**

    .. code-block:: sql

        select
            a,
            *,
            row_number() over (partition by id order by date) as y,
            b
        from x


    **Best practice**

    Order ``select`` targets in ascending complexity

    .. code-block:: sql

        select
            *,
            a,
            b,
            row_number() over (partition by id order by date) as y
        from x

    """

    name = "structure.column_order"
    aliases = ("L034",)
    groups = ("all", "structure")
    crawl_behaviour = SegmentSeekerCrawler({"select_clause"})
    is_fix_compatible = True

    def _is_view_with_explicit_columns(self, context: RuleContext) -> bool:
        """Check if SELECT is in a CREATE VIEW with explicit column list.

        Args:
            context: The rule context containing parent stack

        Returns:
            True if this SELECT is in a CREATE VIEW with explicit columns
        """
        # Traverse parent stack looking for create_view_statement
        for parent in context.parent_stack:
            if parent.is_type("create_view_statement"):
                # Check if the view has an explicit column list
                # Look for a bracketed segment containing column references
                for child in parent.segments:
                    if child.is_type("bracketed"):
                        # Check if this bracketed segment contains column references
                        # ANSI-based dialects (Snowflake, PostgreSQL, etc.)
                        # use column_reference
                        if any(
                            seg.is_type("column_reference")
                            for seg in child.recursive_crawl("column_reference")
                        ):
                            return True
                        # T-SQL dialect uses index_column_definition instead
                        if any(
                            seg.is_type("index_column_definition")
                            for seg in child.recursive_crawl("index_column_definition")
                        ):
                            return True
                        # BigQuery represents the view column list with
                        # column_definition segments instead
                        if any(
                            seg.is_type("column_definition")
                            for seg in child.recursive_crawl("column_definition")
                        ):
                            return True
                        # ClickHouse lists the columns as bare identifiers
                        # directly in the brackets. Only the column list holds
                        # them at the top level -- a parenthesized `AS (SELECT
                        # ...)` body wraps them in a select_statement instead --
                        # so check the bracket's direct children.
                        if any(
                            seg.is_type("naked_identifier", "quoted_identifier")
                            for seg in child.segments
                        ):
                            return True
                # Found a CREATE VIEW but no explicit column list
                return False
        return False

    def _validate(self, i: int, segment: BaseSegment) -> None:
        # Check if we've seen a more complex select target element already
        if self.seen_band_elements[i + 1 : :] != [[]] * len(
            self.seen_band_elements[i + 1 : :]
        ):
            # Found a violation (i.e. a simpler element that *follows* a more
            # complex element.
            self.violation_exists = True
        self.current_element_band: Optional[int] = i
        self.seen_band_elements[i].append(segment)

    @classmethod
    def _is_simple_function(cls, segment: BaseSegment) -> bool:
        """Return whether a function segment is considered simple."""
        function_name = segment.get_child("function_name")
        return bool(function_name and function_name.raw_upper == "CAST")

    @classmethod
    def _is_simple_cast_expression(cls, segment: BaseSegment) -> bool:
        """Return whether a cast expression wraps a simple expression."""
        if not segment.is_type("cast_expression"):
            return False
        return cls._is_simple_expression_segment(segment.segments[0])

    @classmethod
    def _is_simple_expression_segment(cls, segment: BaseSegment) -> bool:
        """Return whether a segment should be treated as a simple expression."""
        if segment.is_type("column_reference", "object_reference", "literal"):
            return True
        if segment.is_type("function"):
            return cls._is_simple_function(segment)
        if segment.is_type("cast_expression"):
            return cls._is_simple_cast_expression(segment)
        return False

    @classmethod
    def _is_simple_expression(cls, segment: BaseSegment, child_type: str) -> bool:
        """Return whether an expression child should be treated as simple."""
        if not segment.is_type("expression") or not segment.get_child(child_type):
            return False
        if len(segment.segments) not in (1, 2):
            return False
        return cls._is_simple_expression_segment(segment.segments[0])

    def _eval(self, context: RuleContext) -> EvalResultType:
        self.violation_exists = False
        # Bands of select targets in order to be enforced
        select_element_order_preference: tuple[
            tuple[Union[str, tuple[str, ...]], ...], ...
        ] = (
            ("wildcard_expression",),
            (
                "object_reference",
                "literal",
                "cast_expression",
                ("function", "cast"),
                ("expression", "cast_expression"),
            ),
        )

        # Track which bands have been seen, with additional empty list for the
        # non-matching elements. If we find a matching target element, we append the
        # element to the corresponding index.
        self.seen_band_elements: list[list[BaseSegment]] = [
            [] for _ in select_element_order_preference
        ] + [[]]  # type: ignore

        assert context.segment.is_type("select_clause")

        # insert, merge, create table, union are order-sensitive
        for seg in reversed(context.parent_stack):
            if seg.is_type(
                "insert_statement",
                "set_expression",
                "create_table_statement",
                "merge_statement",
            ):
                return None

        # CTE is order-sensitive only if CTE is referenced as SELECT * in set expression
        for seg in reversed(context.parent_stack):
            if seg.is_type("common_table_expression"):
                cte_identifier = seg.get_child("identifier")
                assert cte_identifier is not None
                maybe_with_compound_statement = seg.get_parent()
                if maybe_with_compound_statement is None:
                    break  # pragma: no cover
                with_compound_statement, _ = maybe_with_compound_statement
                for ref in with_compound_statement.recursive_crawl("table_reference"):
                    if ref.raw_upper == cte_identifier.raw_upper:
                        path = with_compound_statement.path_to(ref)
                        if any(
                            path_step.segment.is_type("set_expression")
                            for path_step in path
                        ):
                            select_statements = [
                                path_step.segment
                                for path_step in path
                                if path_step.segment.is_type(
                                    "select_statement",
                                    "unordered_select_statement_segment",
                                )
                            ]
                            if any(
                                "wildcard_expression"
                                in select_statement.descendant_type_set
                                for select_statement in select_statements
                            ):
                                return None

        # Skip reordering for CREATE VIEW with explicit column list
        if self._is_view_with_explicit_columns(context):
            return None

        select_clause_segment = context.segment
        select_target_elements = context.segment.get_children("select_clause_element")
        if not select_target_elements:  # pragma: no cover
            return None

        # Iterate through all the select targets to find any order violations
        for segment in select_target_elements:
            # The band index of the current segment in
            # select_element_order_preference
            self.current_element_band = None

            # Compare the segment to the bands in select_element_order_preference
            for i, band in enumerate(select_element_order_preference):
                for e in band:
                    # Identify simple select target
                    if isinstance(e, str) and segment.get_child(e):
                        child = segment.get_child(e)
                        assert child
                        if e != "cast_expression" or self._is_simple_cast_expression(
                            child
                        ):
                            self._validate(i, segment)

                    # Identify function
                    elif isinstance(e, tuple) and e[0] == "function":
                        try:
                            _function = segment.get_child("function")
                            assert _function
                            if self._is_simple_function(_function):
                                self._validate(i, segment)
                        except (AttributeError, AssertionError):
                            # If the segment doesn't match
                            pass

                    # Identify simple expression
                    elif isinstance(e, tuple) and e[0] == "expression":
                        try:
                            _expression = segment.get_child("expression")
                            assert _expression

                            if self._is_simple_expression(_expression, e[1]):
                                self._validate(i, segment)
                        except (AttributeError, AssertionError):
                            # If the segment doesn't match
                            pass

            # If the target doesn't exist in select_element_order_preference then it
            # is 'complex' and must go last
            if self.current_element_band is None:
                self.seen_band_elements[-1].append(segment)

        if self.violation_exists:
            if len(context.parent_stack) and any(
                self._implicit_column_references(context.parent_stack[-1])
            ):
                # If there are implicit column references (i.e. column
                # numbers), warn but don't fix, because it's much more
                # complicated to autofix.
                return LintResult(anchor=select_clause_segment)
            # Create a list of all the edit fixes
            # We have to do this at the end of iterating through all the
            # select_target_elements to get the order correct. This means we can't
            # add a lint fix to each individual LintResult as we go
            ordered_select_target_elements = [
                segment for band in self.seen_band_elements for segment in band
            ]
            # TODO: The "if" in the loop below compares corresponding items
            # to avoid creating "do-nothing" edits. A potentially better
            # approach would leverage difflib.SequenceMatcher.get_opcodes(),
            # which generates a list of edit actions (similar to the
            # command-line "diff" tool in Linux). This is more complex to
            # implement, but minimizing the number of LintFixes makes the
            # final application of patches (in "sqlfluff fix") more robust.
            fixes = [
                LintFix.replace(
                    initial_select_target_element,
                    [replace_select_target_element],
                )
                for initial_select_target_element, replace_select_target_element in zip(
                    select_target_elements, ordered_select_target_elements
                )
                if initial_select_target_element is not replace_select_target_element
            ]
            # Anchoring on the select statement segment ensures that
            # select statements which include macro targets are ignored
            # when ignore_templated_areas is set
            return LintResult(anchor=select_clause_segment, fixes=fixes)

        return None

    @classmethod
    def _implicit_column_references(cls, segment: BaseSegment) -> Iterator[BaseSegment]:
        """Yield any implicit ORDER BY or GROUP BY column references.

        This function was adapted from similar code in AM06.
        """
        _ignore_types: list[str] = ["withingroup_clause", "window_specification"]
        if not segment.is_type(*_ignore_types):  # Ignore Windowing clauses
            if segment.is_type("groupby_clause", "orderby_clause"):
                for seg in segment.segments:
                    if seg.is_type("numeric_literal"):
                        yield segment
            else:
                for seg in segment.segments:
                    yield from cls._implicit_column_references(seg)



"""init file for the parser."""

from sqlfluff.core.parser.grammar import (
    AnyNumberOf,
    AnySetOf,
    Anything,
    Bracketed,
    Conditional,
    Delimited,
    Nothing,
    OneOf,
    OptionallyBracketed,
    OptionallyDelimited,
    Ref,
    Sequence,
)
from sqlfluff.core.parser.lexer import (
    LexerType,
    PyLexer,
    RegexLexer,
    StringLexer,
    get_lexer_class,
)
from sqlfluff.core.parser.markers import PositionMarker
from sqlfluff.core.parser.matchable import Matchable
from sqlfluff.core.parser.parser import Parser
from sqlfluff.core.parser.parsers import (
    MultiStringParser,
    RegexParser,
    StringParser,
    TypedParser,
)
from sqlfluff.core.parser.segments import (
    BaseFileSegment,
    BaseSegment,
    BinaryOperatorSegment,
    BracketedSegment,
    CodeSegment,
    CommentSegment,
    ComparisonOperatorSegment,
    CompositeBinaryOperatorSegment,
    CompositeComparisonOperatorSegment,
    Dedent,
    IdentifierSegment,
    ImplicitIndent,
    Indent,
    KeywordSegment,
    LiteralKeywordSegment,
    LiteralSegment,
    NewlineSegment,
    RawSegment,
    SegmentGenerator,
    SourceFix,
    SymbolSegment,
    UnlexableSegment,
    WhitespaceSegment,
    WordSegment,
)
from sqlfluff.core.parser.types import ParseMode

# Get the appropriate lexer class (PyRsLexer if available, otherwise PyLexer)
Lexer = get_lexer_class()

__all__ = (
    "BaseSegment",
    "SourceFix",
    "BaseFileSegment",
    "BracketedSegment",
    "RawSegment",
    "CodeSegment",
    "UnlexableSegment",
    "CommentSegment",
    "WhitespaceSegment",
    "NewlineSegment",
    "KeywordSegment",
    "SymbolSegment",
    "IdentifierSegment",
    "LiteralSegment",
    "LiteralKeywordSegment",
    "BinaryOperatorSegment",
    "CompositeBinaryOperatorSegment",
    "ComparisonOperatorSegment",
    "CompositeComparisonOperatorSegment",
    "WordSegment",
    "Indent",
    "Dedent",
    "ImplicitIndent",
    "SegmentGenerator",
    "Sequence",
    "OneOf",
    "Delimited",
    "Bracketed",
    "AnyNumberOf",
    "AnySetOf",
    "Ref",
    "Anything",
    "Nothing",
    "OptionallyBracketed",
    "OptionallyDelimited",
    "Conditional",
    "StringParser",
    "MultiStringParser",
    "TypedParser",
    "RegexParser",
    "PositionMarker",
    "Lexer",
    "PyLexer",
    "LexerType",
    "StringLexer",
    "RegexLexer",
    "Parser",
    "Matchable",
    "ParseMode",
)



"""Defines small container classes to hold intermediate results during linting."""

from typing import Any, NamedTuple, Optional, Union

from sqlfluff.core.config import FluffConfig
from sqlfluff.core.errors import (
    SQLBaseError,
    SQLLexError,
    SQLParseError,
    SQLTemplaterError,
)
from sqlfluff.core.parser.segments.base import BaseSegment
from sqlfluff.core.templaters import TemplatedFile


class RuleTuple(NamedTuple):
    """Rule Tuple object for describing rules."""

    code: str
    name: str
    description: str
    groups: tuple[str, ...]
    aliases: tuple[str, ...]


class RenderedFile(NamedTuple):
    """An object to store the result of a templated file/string.

    This is notable as it's the intermediate state between what happens
    in the main process and the child processes when running in parallel mode.
    """

    templated_variants: list[TemplatedFile]
    templater_violations: list[SQLTemplaterError]
    config: FluffConfig
    time_dict: dict[str, float]
    fname: str
    encoding: str
    source_str: str


class DeferredRenderTask(NamedTuple):
    """A minimal task packet sent to workers when templating is deferred.

    Instead of templating in the main process and shipping the full
    RenderedFile (~200 KB) across the IPC boundary, we ship only the
    filename and root config and let the worker call render_file itself.
    """

    fname: str
    root_config: FluffConfig
    fix: bool


class ParsedVariant(NamedTuple):
    """An object to store the result of parsing a single TemplatedFile.

    Args:
        templated_file (:obj:`TemplatedFile`): Containing the details
            of the templated file. If templating fails, this will be `None`.
        tree (:obj:`BaseSegment`): The segment structure representing the
            parsed file. If parsing fails due to an unrecoverable
            violation then we will be None.
        lexing_violations (:obj:`list` of :obj:`SQLLexError`): Any violations
            raised during the lexing phase.
        parsing_violations (:obj:`list` of :obj:`SQLParseError`): Any violations
            raised during the parsing phase.
    """

    templated_file: TemplatedFile
    tree: Optional[BaseSegment]
    lexing_violations: list[SQLLexError]
    parsing_violations: list[SQLParseError]

    def violations(self) -> list[Union[SQLLexError, SQLParseError]]:
        """Returns the combined lexing and parsing violations for this variant."""
        return [*self.lexing_violations, *self.parsing_violations]


class ParsedString(NamedTuple):
    """An object to store the result of parsing a string.

    Args:
        parsed_variants (:obj:`list` of :obj:`ParsedVariant`): The parsed
            variants of this file. Empty if parsing or templating failed.
        templating_violations (:obj:`list` of :obj:`SQLTemplaterError`):
            Any violations raised during the templating phase. Any violations
            raised during lexing or parsing can be found in the
            `parsed_variants`, or accessed using the `.violations()` method
            which combines all the violations.
        time_dict (:obj:`dict`): Contains timings for how long each step
            took in the process.
        config (:obj:`FluffConfig`): The active config for this file,
            including any parsed in-file directives.
        fname (str): The name of the file. Used mostly for user feedback.
        source_str (str): The raw content of the source file.
    """

    parsed_variants: list[ParsedVariant]
    templating_violations: list[SQLTemplaterError]
    time_dict: dict[str, Any]
    config: FluffConfig
    fname: str
    source_str: str

    @property
    def violations(self) -> list[SQLBaseError]:
        """Returns the combination of violations for this variant.

        NOTE: This is implemented as a property for backward compatibility.
        """
        root_variant = self.root_variant()
        if root_variant:
            return [*self.templating_violations, *root_variant.violations()]
        return [
            *self.templating_violations,
            *(v for variant in self.parsed_variants for v in variant.violations()),
        ]

    def root_variant(self) -> Optional[ParsedVariant]:
        """Returns the root variant if successfully parsed, otherwise None."""
        if not self.parsed_variants:
            # In the case of a fatal templating error, there will be no valid
            # variants. Return None.
            return None
        root_variant = self.parsed_variants[0]
        if not root_variant.tree:
            # In the case of a parsing fail, there will be a variant, but it will
            # have failed to parse and so will have a null tree. Count this as
            # an inappropriate variant to return, so return None.
            return None
        return root_variant

    @property
    def tree(self) -> BaseSegment:
        """Return the main variant tree.

        NOTE: This method is primarily for testing convenience and therefore
        asserts that parsing has been successful. If this isn't appropriate
        for the given use case, then don't use this property.
        """
        assert self.parsed_variants, "No successfully parsed variants."
        root_variant = self.parsed_variants[0]
        assert root_variant.tree, "Root variant not successfully parsed."
        return root_variant.tree



"""Information API."""

from sqlfluff.core import Linter, dialect_readout
from sqlfluff.core.dialects import DialectTuple
from sqlfluff.core.linter import RuleTuple


def list_rules() -> list[RuleTuple]:
    """Return a list of available rule tuples."""
    linter = Linter()
    return linter.rule_tuples()


def list_dialects() -> list[DialectTuple]:
    """Return a list of available dialect info."""
    return list(dialect_readout())
